# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""init trove db

Revision ID: 906cffda7b29
Revises:
Create Date: 2024-04-30 13:58:12.700444
"""

from typing import Sequence, Union

from alembic import op
from sqlalchemy import Boolean, Column, DateTime, Float, ForeignKey, Integer, \
    String, Text, UniqueConstraint
from sqlalchemy.sql import table, column
from trove.common import cfg

CONF = cfg.CONF


# revision identifiers, used by Alembic.
revision: str = '906cffda7b29'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    bind = op.get_bind()

    """ merges the schmas up to 048 are treated asis.
    049 is saved as another revision
    """

    """001_base_schema.py
    """
    # NOTE(hiwkby)
    # move the 001_base_schema.py to 032_clusters.py because
    # instances references foreign keys in clusters and other tables
    # that are created at 032_clusters.py.

    """002_service_images.py
    """
    op.create_table(
        'service_images',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('service_name', String(255)),
        Column('image_id', String(255))
    )

    """003_service_statuses.py
    """
    op.create_table(
        'service_statuses',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('instance_id', String(36), nullable=False),
        Column('status_id', Integer(), nullable=False),
        Column('status_description', String(64), nullable=False),
        Column('updated_at', DateTime()),
    )
    op.create_index('service_statuses_instance_id', 'service_statuses',
                    ['instance_id'])

    """004_root_enabled.py
    """
    op.create_table(
        'root_enabled_history',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('user', String(length=255)),
        Column('created', DateTime()),
    )

    """005_heartbeat.py
    """
    op.create_table(
        'agent_heartbeats',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('instance_id', String(36), nullable=False, index=True,
               unique=True),
        Column('guest_agent_version', String(255), index=True),
        Column('deleted', Boolean(), index=True),
        Column('deleted_at', DateTime()),
        Column('updated_at', DateTime(), nullable=False)
    )

    """006_dns_records.py
    """
    op.create_table(
        'dns_records',
        Column('name', String(length=255), primary_key=True),
        Column('record_id', String(length=64))
    )

    """007_add_volume_flavor.py
    """
    # added the columns to instances table

    """008_add_instance_fields.py
    """
    # added the columns to instances table

    """009_add_deleted_flag_to_instances.py
    """
    # added the columns to instances table

    """010_add_usage.py
    """
    op.create_table(
        'usage_events',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('instance_name', String(36)),
        Column('tenant_id', String(36)),
        Column('nova_instance_id', String(36)),
        Column('instance_size', Integer()),
        Column('nova_volume_id', String(36)),
        Column('volume_size', Integer()),
        Column('end_time', DateTime()),
        Column('updated', DateTime()),
    )

    """011_quota.py
    """
    op.create_table(
        'quotas',
        Column('id', String(36),
               primary_key=True, nullable=False),
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('tenant_id', String(36)),
        Column('resource', String(length=255), nullable=False),
        Column('hard_limit', Integer()),
        UniqueConstraint('tenant_id', 'resource')
    )

    op.create_table(
        'quota_usages',
        Column('id', String(36),
               primary_key=True, nullable=False),
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('tenant_id', String(36)),
        Column('in_use', Integer(), default=0),
        Column('reserved', Integer(), default=0),
        Column('resource', String(length=255), nullable=False),
        UniqueConstraint('tenant_id', 'resource')
    )

    op.create_table(
        'reservations',
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('id', String(36),
               primary_key=True, nullable=False),
        Column('usage_id', String(36)),
        Column('delta', Integer(), nullable=False),
        Column('status', String(length=36))
    )

    """012_backup.py
    """
    # NOTE(hiwkby)
    # move the 012_backup.py to 029_add_backup_datastore.py because
    # backups references a datastore_versions.id as a foreign key
    # that are created at 012_backup.py.

    """013_add_security_group_artifacts.py
    """
    op.create_table(
        'security_groups',
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('name', String(length=255)),
        Column('description', String(length=255)),
        Column('user', String(length=255)),
        Column('tenant_id', String(length=255)),
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('deleted', Boolean(), default=0),
        Column('deleted_at', DateTime()),
    )

    # NOTE(hiwkby)
    # move the security_group_instance_associations schema
    # to 032_clusters.py where instances table is created.

    op.create_table(
        'security_group_rules',
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('group_id', String(length=36),
               ForeignKey('security_groups.id', ondelete='CASCADE',
                          onupdate='CASCADE')),
        Column('parent_group_id', String(length=36),
               ForeignKey('security_groups.id', ondelete='CASCADE',
                          onupdate='CASCADE')),
        Column('protocol', String(length=255)),
        Column('from_port', Integer()),
        Column('to_port', Integer()),
        Column('cidr', String(length=255)),
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('deleted', Boolean(), default=0),
        Column('deleted_at', DateTime()),
    )

    """014_update_instance_flavor_id.py
    """
    # updated the instances table schema

    """015_add_service_type.py
    """
    # updated the instances table schema
    # NOTE(hiwkby)
    # service_type was deleted in 016_add_datastore_type.py

    """016_add_datastore_type.py
    """
    op.create_table(
        'datastores',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255), unique=True),
        # NOTE(hiwkby) manager was dropped in 017_update_datastores.py
        # Column('manager', String(255), nullable=False),
        Column('default_version_id', String(36)),
    )
    op.create_table(
        'datastore_versions',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('datastore_id', String(36), ForeignKey('datastores.id')),
        # NOTE(hiwkby)
        # unique=false in 018_datastore_versions_fix.py
        Column('name', String(255), unique=False),
        Column('image_id', String(36), nullable=True),
        Column('packages', String(511)),
        Column('active', Boolean(), nullable=False),
        # manager added in 017_update_datastores.py
        Column('manager', String(255)),
        # image_tags added in 047_image_tag_in_datastore_version.py
        Column('image_tags', String(255), nullable=True),
        # version added in 048_add_version_to_datastore_version.py
        Column('version', String(255), nullable=True),
        UniqueConstraint('datastore_id', 'name', 'version',
                         name='ds_versions')
    )

    """017_update_datastores.py
    """
    # updated the datastores and datastore_versions table schema.

    """018_datastore_versions_fix.py
    """
    # updated the datastore_versions table schema

    """019_datastore_fix.py
    """
    # updated the datastore_versions table schema

    """020_configurations.py
    """
    op.create_table(
        'configurations',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(64), nullable=False),
        Column('description', String(256)),
        Column('tenant_id', String(36), nullable=False),
        Column('datastore_version_id', String(36), nullable=False),
        Column('deleted', Boolean(), nullable=False, default=False),
        Column('deleted_at', DateTime()),
        # NOTE(hiwkby)
        # created added in 031_add_timestamps_to_configurations.py
        Column('created', DateTime()),
        Column('updated', DateTime()),
    )

    op.create_table(
        'configuration_parameters',
        Column('configuration_id', String(36), ForeignKey('configurations.id'),
               nullable=False, primary_key=True),
        Column('configuration_key', String(128),
               nullable=False, primary_key=True),
        Column('configuration_value', String(128)),
        Column('deleted', Boolean(), nullable=False, default=False),
        Column('deleted_at', DateTime()),
    )

    """021_conductor_last_seen.py
    """
    op.create_table(
        'conductor_lastseen',
        Column('instance_id', String(36), primary_key=True, nullable=False),
        Column('method_name', String(36), primary_key=True, nullable=False),
        Column('sent', Float(precision=32))
    )

    """022_add_backup_parent_id.py
    """
    # updated the backups table schema

    """023_add_instance_indexes.py
    """
    # updated the instances table schema

    """024_add_backup_indexes.py
    """
    # updated the backups table schema

    """025_add_service_statuses_indexes.py
    """
    # updated the service_statuses table schema

    """026_datastore_versions_unique_fix.py
    """
    # updated the datastore_versions table schema

    """027_add_datastore_capabilities.py
    """
    op.create_table(
        'capabilities',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255), unique=True),
        Column('description', String(255), nullable=False),
        Column('enabled', Boolean()),
    )

    op.create_table(
        'capability_overrides',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('datastore_version_id', String(36),
               ForeignKey('datastore_versions.id')),
        Column('capability_id', String(36), ForeignKey('capabilities.id')),
        Column('enabled', Boolean()),
        UniqueConstraint('datastore_version_id', 'capability_id',
                         name='idx_datastore_capabilities_enabled')
    )

    """028_recreate_agent_heartbeat.py
    """
    # updated the datastore_versions table schema

    """029_add_backup_datastore.py
    """
    # NOTE(hiwkby)
    # moves here from 012_backup.py because
    # backups references datastore_versions.id as a foreign key
    op.create_table(
        'backups',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255), nullable=False),
        Column('description', String(512)),
        Column('location', String(1024)),
        Column('backup_type', String(32)),
        Column('size', Float()),
        Column('tenant_id', String(36)),
        Column('state', String(32), nullable=False),
        Column('instance_id', String(36)),
        Column('checksum', String(32)),
        Column('backup_timestamp', DateTime()),
        Column('deleted', Boolean(), default=0),
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('deleted_at', DateTime()),
        # 022_add_backup_parent_id.py
        Column('parent_id', String(36), nullable=True),
        # 029_add_backup_datastore.py
        Column('datastore_version_id', String(36),
               ForeignKey(
                   "datastore_versions.id",
                   name="backups_ibfk_1")),
    )
    op.create_index('backups_instance_id', 'backups', ['instance_id'])
    op.create_index('backups_deleted', 'backups', ['deleted'])

    """030_add_master_slave.py
    """
    # updated the instances table schema

    """031_add_timestamps_to_configurations.py
    """
    # updated the configurations table schema

    """032_clusters.py
    """
    op.create_table(
        'clusters',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('created', DateTime(), nullable=False),
        Column('updated', DateTime(), nullable=False),
        Column('name', String(255), nullable=False),
        Column('task_id', Integer(), nullable=False),
        Column('tenant_id', String(36), nullable=False),
        Column('datastore_version_id', String(36),
               ForeignKey(
                   "datastore_versions.id",
                   name="clusters_ibfk_1"),
               nullable=False),
        Column('deleted', Boolean()),
        Column('deleted_at', DateTime()),
        Column('configuration_id', String(36),
               ForeignKey(
                   "configurations.id",
                   name="clusters_ibfk_2")),
    )

    op.create_index('clusters_tenant_id', 'clusters', ['tenant_id'])
    op.create_index('clusters_deleted', 'clusters', ['deleted'])

    # NOTE(hiwkby)
    # move here from the 001_base_schema.py because
    # instances references cluster and other tables as foreign keys
    op.create_table(
        'instances',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('name', String(255)),
        Column('hostname', String(255)),
        Column('compute_instance_id', String(36)),
        Column('task_id', Integer()),
        Column('task_description', String(255)),
        Column('task_start_time', DateTime()),
        Column('volume_id', String(36)),
        Column('flavor_id', String(255)),
        Column('volume_size', Integer()),
        Column('tenant_id', String(36), nullable=True),
        Column('server_status', String(64)),
        Column('deleted', Boolean()),
        Column('deleted_at', DateTime()),
        Column('datastore_version_id', String(36),
               ForeignKey(
                   "datastore_versions.id",
                   name="instances_ibfk_1"), nullable=True),
        Column('configuration_id', String(36),
               ForeignKey(
                   "configurations.id",
                   name="instances_ibfk_2")),
        Column('slave_of_id', String(36),
               ForeignKey(
                   "instances.id",
                   name="instances_ibfk_3"), nullable=True),
        Column('cluster_id', String(36),
               ForeignKey(
                   "clusters.id",
                   name="instances_ibfk_4")),
        Column('shard_id', String(36)),
        Column('type', String(64)),
        Column('region_id', String(255)),
        Column('encrypted_key', String(255)),
        Column('access', Text(), nullable=True),
    )
    op.create_index('instances_tenant_id', 'instances', ['tenant_id'])
    op.create_index('instances_deleted', 'instances', ['deleted'])
    op.create_index('instances_cluster_id', 'instances', ['cluster_id'])

    # NOTE(hiwkby)
    # move here from the security_group_instance_associations schema
    # because it references instances.id as a foreign key
    op.create_table(
        'security_group_instance_associations',
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('security_group_id', String(length=36),
               ForeignKey('security_groups.id', ondelete='CASCADE',
                          onupdate='CASCADE')),
        Column('instance_id', String(length=36),
               ForeignKey('instances.id', ondelete='CASCADE',
                          onupdate='CASCADE')),
        Column('created', DateTime()),
        Column('updated', DateTime()),
        Column('deleted', Boolean(), default=0),
        Column('deleted_at', DateTime())
    )

    """033_datastore_parameters.py
    """
    op.create_table(
        'datastore_configuration_parameters',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(128), primary_key=True, nullable=False),
        Column('datastore_version_id', String(36),
               ForeignKey("datastore_versions.id"),
               primary_key=True, nullable=False),
        Column('restart_required', Boolean(), nullable=False, default=False),
        Column('max_size', String(40)),
        Column('min_size', String(40)),
        Column('data_type', String(128), nullable=False),
        UniqueConstraint(
            'datastore_version_id', 'name',
            name='UQ_datastore_configuration_parameters_datastore_version_id_name')  # noqa
    )

    """034_change_task_description.py
    """
    # updated the configurations table schema

    """035_flavor_id_int_to_string.py
    """
    # updated the configurations table schema

    """036_add_datastore_version_metadata.py
    """
    op.create_table(
        'datastore_version_metadata',
        Column('id', String(36), primary_key=True, nullable=False),
        Column(
            'datastore_version_id',
            String(36),
            ForeignKey('datastore_versions.id', ondelete='CASCADE'),
        ),
        Column('key', String(128), nullable=False),
        Column('value', String(128)),
        Column('created', DateTime(), nullable=False),
        Column('deleted', Boolean(), nullable=False, default=False),
        Column('deleted_at', DateTime()),
        Column('updated_at', DateTime()),
        UniqueConstraint(
            'datastore_version_id', 'key', 'value',
            name='UQ_datastore_version_metadata_datastore_version_id_key_value') # noqa
    )

    """037_modules.py
    """
    is_nullable = True if bind.engine.name == "sqlite" else False
    op.create_table(
        'modules',
        Column('id', String(length=64), primary_key=True, nullable=False),
        Column('name', String(length=255), nullable=False),
        Column('type', String(length=255), nullable=False),
        # contents was updated in 40_module_priority.py
        Column('contents', Text(length=4294967295), nullable=False),
        Column('description', String(length=255)),
        Column('tenant_id', String(length=64), nullable=True),
        Column('datastore_id', String(length=64), nullable=True),
        Column('datastore_version_id', String(length=64), nullable=True),
        Column('auto_apply', Boolean(), default=0, nullable=False),
        Column('visible', Boolean(), default=1, nullable=False),
        Column('live_update', Boolean(), default=0, nullable=False),
        Column('md5', String(length=32), nullable=False),
        Column('created', DateTime(), nullable=False),
        Column('updated', DateTime(), nullable=False),
        Column('deleted', Boolean(), default=0, nullable=False),
        Column('deleted_at', DateTime()),
        UniqueConstraint(
            'type', 'tenant_id', 'datastore_id', 'datastore_version_id',
            'name', 'deleted_at',
            name='UQ_type_tenant_datastore_datastore_version_name'),
        # NOTE(hiwkby)
        # the following columns added in 040_module_priority.py
        Column('priority_apply', Boolean(), nullable=is_nullable, default=0),
        Column('apply_order', Integer(), nullable=is_nullable, default=5),
        Column('is_admin', Boolean(), nullable=is_nullable, default=0),
    )

    op.create_table(
        'instance_modules',
        Column('id', String(length=64), primary_key=True, nullable=False),
        Column('instance_id', String(length=64),
               ForeignKey('instances.id', ondelete="CASCADE",
                          onupdate="CASCADE"), nullable=False),
        Column('module_id', String(length=64),
               ForeignKey('modules.id', ondelete="CASCADE",
                          onupdate="CASCADE"), nullable=False),
        Column('md5', String(length=32), nullable=False),
        Column('created', DateTime(), nullable=False),
        Column('updated', DateTime(), nullable=False),
        Column('deleted', Boolean(), default=0, nullable=False),
        Column('deleted_at', DateTime()),
    )

    """038_instance_faults.py
    """
    op.create_table(
        'instance_faults',
        Column('id', String(length=64), primary_key=True, nullable=False),
        Column('instance_id', String(length=64),
               ForeignKey('instances.id', ondelete="CASCADE",
                          onupdate="CASCADE"), nullable=False),
        Column('message', String(length=255), nullable=False),
        Column('details', Text(length=65535), nullable=False),
        Column('created', DateTime(), nullable=False),
        Column('updated', DateTime(), nullable=False),
        Column('deleted', Boolean(), default=0, nullable=False),
        Column('deleted_at', DateTime()),
    )

    """039_region.py
    """
    instances = table("instances", column("region_id", String))
    op.execute(
        instances.update()
        .values({"region_id": CONF.service_credentials.region_name})
    )

    """040_module_priority.py
    """
    # updated the modules table schema

    """041_instance_keys.py
    """
    # updated the instances table schema

    """042_add_cluster_configuration_id.py
    """
    # updated the clusters table schema

    """043_instance_ds_version_nullable.py
    """
    # updated the instances table schema

    """044_remove_datastore_configuration_parameters_deleted.py
    """
    # updated the datastore_configuration_parameters table schema

    """045_add_backup_strategy.py
    """
    op.create_table(
        'backup_strategy',
        Column('id', String(36), primary_key=True, nullable=False),
        Column('tenant_id', String(36), nullable=False),
        Column('instance_id', String(36), nullable=False, default=''),
        Column('backend', String(255), nullable=False),
        Column('swift_container', String(255), nullable=True),
        Column('created', DateTime()),
        UniqueConstraint(
            'tenant_id', 'instance_id',
            name='UQ_backup_strategy_tenant_id_instance_id'),
    )
    op.create_index('backup_strategy_tenant_id_instance_id',
                    'backup_strategy', ['tenant_id', 'instance_id'])

    """046_add_access_to_instance.py
    """
    # updated the instances table schema

    """047_image_tag_in_datastore_version.py
    """
    # updated the datastore_versions table schema

    """048_add_version_to_datastore_version.py
    """
    # updated the datastore_versions table schema

    """049_add_registry_ext_to_datastore_version.py
    """


def downgrade() -> None:
    pass



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import os.path

from oslo_log import log as logging

from trove.common import cfg
from trove.common import configurations
from trove.common import constants
from trove.common import exception
from trove.common.notification import EndNotification
from trove.common import utils
from trove.guestagent.common import operating_system
from trove.guestagent.datastore import manager
from trove.guestagent import guest_log
from trove.guestagent.utils import docker as docker_util
from trove.guestagent.utils import mysql as mysql_util
from trove.instance import service_status

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class MySqlManager(manager.Manager):
    def __init__(self, mysql_app, mysql_app_status, mysql_admin,
                 manager_name='mysql'):
        super(MySqlManager, self).__init__(manager_name)

        self.app = mysql_app
        self.status = mysql_app_status
        self.adm = mysql_admin
        self.volume_do_not_start_on_reboot = False

    @property
    def configuration_manager(self):
        return self.app.configuration_manager

    def get_service_status(self):
        try:
            with mysql_util.SqlClient(self.app.get_engine()) as client:
                cmd = "SELECT 1;"
                client.execute(cmd)

            LOG.debug("Database service check: database query is responsive")
            return service_status.ServiceStatuses.HEALTHY
        except Exception:
            return super(MySqlManager, self).get_service_status()

    def get_start_db_params(self, data_dir):
        return f'--datadir={data_dir}'

    def do_prepare(self, context, packages, databases, memory_mb, users,
                   device_path, mount_point, backup_info,
                   config_contents, root_password, overrides,
                   cluster_config, snapshot, ds_version=None):
        """This is called from prepare in the base class."""
        data_dir = mount_point + '/data'
        self.app.stop_db()
        operating_system.ensure_directory(
            data_dir,
            user=self.app.database_service_uid,
            group=self.app.database_service_gid,
            as_root=True)
        # This makes sure the include dir is created.
        self.app.set_data_dir(data_dir)

        # Prepare mysql configuration
        LOG.info('Preparing database configuration')
        self.app.configuration_manager.reset_configuration(config_contents)
        self.app.update_overrides(overrides)

        # Restore data from backup and reset root password
        if backup_info:
            self.perform_restore(context, data_dir, backup_info)
            self.reset_password_for_restore(ds_version=ds_version,
                                            data_dir=data_dir)

        # Start database service.
        command = self.get_start_db_params(data_dir)
        self.app.start_db(ds_version=ds_version, command=command)

        self.app.secure()
        enable_remote_root = (backup_info and self.adm.is_root_enabled())
        if enable_remote_root:
            self.status.report_root(context)
        else:
            self.app.secure_root()

        if snapshot:
            # This instance is a replication slave
            self.attach_replica(context, snapshot, snapshot['config'])

    def create_backup(self, context, backup_info):
        """Create backup for the database.

        :param context: User context object.
        :param backup_info: a dictionary containing the db instance id of the
                            backup task, location, type, and other data.
        """
        LOG.info(f"Creating backup {backup_info['id']}")
        with EndNotification(context):
            # Set /var/run/mysqld to allow localhost access.
            volumes_mapping = {
                '/var/lib/mysql': {'bind': '/var/lib/mysql', 'mode': 'rw'},
                constants.MYSQL_HOST_SOCKET_PATH: {"bind": "/var/run/mysqld",
                                                   "mode": "ro"},
                '/tmp': {'bind': '/tmp', 'mode': 'rw'}
            }
            self.app.create_backup(context, backup_info,
                                   volumes_mapping=volumes_mapping,
                                   need_dbuser=True)

    def get_datastore_log_defs(self):
        owner = cfg.get_configuration_property('database_service_uid')
        datastore_dir = self.app.get_data_dir()
        server_section = configurations.MySQLConfParser.SERVER_CONF_SECTION
        long_query_time = CONF.get(self.manager).get(
            'guest_log_long_query_time') / 1000
        general_log_file = self.build_log_file_name(
            self.GUEST_LOG_DEFS_GENERAL_LABEL, owner,
            datastore_dir=datastore_dir)
        error_log_file = self.validate_log_file('/var/log/mysqld.log', owner)
        slow_query_log_file = self.build_log_file_name(
            self.GUEST_LOG_DEFS_SLOW_QUERY_LABEL, owner,
            datastore_dir=datastore_dir)
        return {
            self.GUEST_LOG_DEFS_GENERAL_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.USER,
                self.GUEST_LOG_USER_LABEL: owner,
                self.GUEST_LOG_FILE_LABEL: general_log_file,
                self.GUEST_LOG_SECTION_LABEL: server_section,
                self.GUEST_LOG_ENABLE_LABEL: {
                    'general_log': 'on',
                    'general_log_file': general_log_file,
                    'log_output': 'file',
                },
                self.GUEST_LOG_DISABLE_LABEL: {
                    'general_log': 'off',
                },
            },
            self.GUEST_LOG_DEFS_SLOW_QUERY_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.USER,
                self.GUEST_LOG_USER_LABEL: owner,
                self.GUEST_LOG_FILE_LABEL: slow_query_log_file,
                self.GUEST_LOG_SECTION_LABEL: server_section,
                self.GUEST_LOG_ENABLE_LABEL: {
                    'slow_query_log': 'on',
                    'slow_query_log_file': slow_query_log_file,
                    'long_query_time': long_query_time,
                },
                self.GUEST_LOG_DISABLE_LABEL: {
                    'slow_query_log': 'off',
                },
            },
            self.GUEST_LOG_DEFS_ERROR_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.SYS,
                self.GUEST_LOG_USER_LABEL: owner,
                self.GUEST_LOG_FILE_LABEL: error_log_file,
            },
        }

    def is_log_enabled(self, logname):
        if logname == self.GUEST_LOG_DEFS_GENERAL_LABEL:
            value = self.configuration_manager.get_value(
                'general_log', section='mysqld', default='off')
            LOG.debug(f"The config value of general_log is {value}")
            return value == 'on'
        elif logname == self.GUEST_LOG_DEFS_SLOW_QUERY_LABEL:
            value = self.configuration_manager.get_value(
                'slow_query_log', section='mysqld', default='off')
            LOG.debug(f"The config value of slow_query_log is {value}")
            return value == 'on'

        return False

    def apply_overrides(self, context, overrides):
        LOG.info("Applying database config.")
        self.app.apply_overrides(overrides)
        LOG.info("Finished applying database config.")

    def reset_password_for_restore(self, ds_version=None,
                                   data_dir='/var/lib/mysql/data'):
        """Reset the root password after restore the db data.

        We create a temporary database container by running mysqld_safe to
        reset the root password.
        """
        LOG.info('Starting to reset password for restore')

        try:
            root_pass = self.app.get_auth_password(file="root.cnf")
        except exception.UnprocessableEntity:
            root_pass = utils.generate_random_password()
            self.app.save_password('root', root_pass)

        init_file = os.path.join(data_dir, "init.sql")
        operating_system.write_file(
            init_file,
            f"ALTER USER 'root'@'localhost' IDENTIFIED BY '{root_pass}';",
            as_root=True
        )
        # Change ownership so the database service user
        # can read it in the container
        operating_system.chown(
            init_file,
            user=self.app.database_service_uid,
            group=self.app.database_service_gid,
            as_root=True
        )

        command = (
            f'mysqld --init-file={init_file} '
            f'--datadir={data_dir} '
        )

        # Start the database container process.
        try:
            self.app.start_db(ds_version=ds_version, command=command)
        except Exception as err:
            LOG.error('Failed to reset password for restore, error: %s',
                      str(err))
            raise err  # re-raised at the end of the finally clause
        finally:
            try:
                LOG.debug(
                    'The init container log: %s',
                    docker_util.get_container_logs(self.app.docker_client)
                )
                docker_util.remove_container(self.app.docker_client)
                # Remove init.sql file after password reset
                operating_system.remove(init_file, force=True, as_root=True)
            except Exception as err:
                LOG.error('Failed to remove container or init file. error: %s',
                          str(err))
                pass
        LOG.info('Finished to reset password for restore')

    def _validate_slave_for_replication(self, context, replica_info):
        if replica_info['replication_strategy'] != self.replication_strategy:
            raise exception.IncompatibleReplicationStrategy(
                replica_info.update({
                    'guest_strategy': self.replication_strategy
                }))

        volume_stats = self.get_filesystem_stats(context, None)
        if (volume_stats.get('total', 0.0) <
                replica_info['dataset']['dataset_size']):
            raise exception.InsufficientSpaceForReplica(
                replica_info.update({
                    'slave_volume_size': volume_stats.get('total', 0.0)
                }))

    def attach_replica(self, context, replica_info, slave_config, **kwargs):
        LOG.info("Attaching replica, replica_info: %s", replica_info)
        try:
            if 'replication_strategy' in replica_info:
                self._validate_slave_for_replication(context, replica_info)

            self.replication.enable_as_slave(self.app, replica_info,
                                             slave_config)
        except Exception as err:
            LOG.error("Error enabling replication, error: %s", str(err))
            self.status.set_status(service_status.ServiceStatuses.FAILED)
            raise

    def make_read_only(self, context, read_only):
        LOG.info("Executing make_read_only(%s)", read_only)
        self.app.make_read_only(read_only)

    def get_latest_txn_id(self, context):
        LOG.info("Calling get_latest_txn_id.")
        return self.app.get_latest_txn_id()

    def get_last_txn(self, context):
        LOG.info("Calling get_last_txn")
        return self.app.get_last_txn()

    def wait_for_txn(self, context, txn):
        LOG.info("Calling wait_for_txn.")
        self.app.wait_for_txn(txn)

    def upgrade(self, context, upgrade_info):
        """Upgrade the database."""
        LOG.info('Starting to upgrade database, upgrade_info: %s',
                 upgrade_info)
        self.app.upgrade(upgrade_info)

    def rebuild(self, context, ds_version, config_contents=None,
                config_overrides=None):
        """Restore datastore service after instance rebuild."""
        LOG.info("Starting to restore database service")
        self.status.begin_install()

        mount_point = CONF.get(CONF.datastore_manager).mount_point
        data_dir = mount_point + '/data'
        operating_system.ensure_directory(data_dir,
                                          user=self.app.database_service_uid,
                                          group=self.app.database_service_gid,
                                          as_root=True)
        # This makes sure the include dir is created.
        self.app.set_data_dir(data_dir)

        try:
            # Prepare mysql configuration
            LOG.debug('Preparing database configuration')
            self.app.configuration_manager.reset_configuration(config_contents)
            self.app.update_overrides(config_overrides)

            # Start database service.
            command = self.get_start_db_params(data_dir)
            self.app.start_db(ds_version=ds_version, command=command)
        except Exception as e:
            LOG.error(f"Failed to restore database service after rebuild, "
                      f"error: {str(e)}")
            self.prepare_error = True
            raise
        finally:
            self.status.end_install(error_occurred=self.prepare_error)

    def post_create_backup(self, context, **kwargs):
        LOG.info("Running post_create_backup")
        try:
            self.app.execute_sql("UNLOCK TABLES;")

            mount_point = CONF.get(CONF.datastore_manager).mount_point
            operating_system.fsunfreeze(mount_point)
        except Exception as e:
            LOG.error("Run post_create_backup failed, error: %s" % str(e))

        return {}



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import hashlib
import io
import json
import tempfile

from keystoneauth1.identity import v3
from keystoneauth1 import session
from oslo_config import cfg
from oslo_log import log as logging
import swiftclient
from swiftclient import exceptions as swift_exc

from backup.storage import base

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def _get_user_keystone_session(auth_url, token, tenant_id):
    auth = v3.Token(
        auth_url=auth_url, token=token,
        project_domain_name="Default",
        project_id=tenant_id
    )
    return session.Session(auth=auth, verify=False)


def _get_service_client(auth_url, token, tenant_id, region_name=None):
    sess = _get_user_keystone_session(auth_url, token, tenant_id)
    os_options = None
    if region_name:
        os_options = {
            'region_name': region_name
        }
    return swiftclient.Connection(session=sess,
                                  os_options=os_options,
                                  insecure=True)


def _set_attr(original):
    """Return a swift friendly header key."""
    key = original.replace('_', '-')
    return 'X-Object-Meta-%s' % key


def _get_attr(original):
    """Get a friendly name from an object header key."""
    key = original.replace('-', '_')
    key = key.replace('x_object_meta_', '')
    return key


class StreamReader(object):
    """Wrap the stream from the backup process and chunk it into segments.
    This class now buffers each segment to a temporary file, making it seekable
    for retry mechanisms, like the one in swiftclient.
    """

    def __init__(self, stream, container, filename, max_file_size):
        self.stream = stream
        self.container = container
        self.filename = filename
        self.max_file_size = max_file_size
        # Will be incremented to 0 in _start_new_segment
        self.file_number = 0
        # True if the entire original stream is exhausted
        self.end_of_file = False
        # tempfile.TemporaryFile for current segment
        self._current_segment_buffer = None
        self._current_segment_checksum = hashlib.md5()
        self._current_segment_length = 0
        # Current read position within _current_segment_buffer
        self._buffer_read_offset = 0
        # True when the current segment is fully read from original stream
        self._segment_fully_buffered = False
        self._prepare_for_new_segment = True

    def _start_new_segment(self):
        # Prepares for a new segment by creating a new buffer and
        #  resetting state.
        if self._current_segment_buffer:
            # Store the final checksum before closing the buffer
            self._current_segment_buffer.close()

        # Using a temporary file for buffering as segments can be large (2GB)
        self._current_segment_buffer = tempfile.TemporaryFile()
        self._current_segment_checksum = hashlib.md5()
        self._current_segment_length = 0
        # Reset read pointer for the new buffer
        self._buffer_read_offset = 0
        # # Reset flag for new segment
        self._segment_fully_buffered = False

    @property
    def base_filename(self):
        """Filename with extensions removed."""
        return self.filename.split('.')[0]

    @property
    def segment(self):
        return '%s_%08d' % (self.base_filename, self.file_number)

    @property
    def first_segment(self):
        return '%s_%08d' % (self.base_filename, 0)

    @property
    def segment_path(self):
        return '%s/%s' % (self.container, self.segment)

    def read(self, chunk_size=2 ** 16):
        """Read data from the stream. This method buffers the current segment
        from the underlying stream to a temporary file, then serves chunks
        from that buffer. This makes the stream seekable.
        """
        # Phase 1: Ensure the current segment is fully buffered from the
        # original stream. This loop will run until the current segment's
        # data is entirely in _current_segment_buffer or the original
        # stream is exhausted.
        if self._prepare_for_new_segment:
            self._start_new_segment()
            self._prepare_for_new_segment = False

        if not self._segment_fully_buffered:
            while True:
                if (self._current_segment_length + chunk_size) > \
                        self.max_file_size:
                    self._segment_fully_buffered = True
                    LOG.info("StreamReader: Current segment %s reached max "
                             "size. Fully buffered.", self.segment)
                    break

                # Read from the original, unseekable stream
                chunk = self.stream.read(chunk_size)

                if not chunk:
                    # Original stream exhausted. Mark overall end of file.
                    self.end_of_file = True
                    self._segment_fully_buffered = True
                    LOG.info("StreamReader: Original stream exhausted. "
                             "Current segment %s fully buffered.",
                             self.segment)
                    break

                self._current_segment_buffer.write(chunk)
                self._current_segment_checksum.update(chunk)
                self._current_segment_length += len(chunk)

            # After buffering is complete for this segment, rewind the buffer
            # so that subsequent reads (by swiftclient,
            # potentially after seek) start from the beginning of the
            # buffered data.
            self._current_segment_buffer.seek(0)
            self._buffer_read_offset = 0
        # Phase 2: Serve data from the buffered temporary file
        data = self._current_segment_buffer.read(chunk_size)
        self._buffer_read_offset += len(data)
        # start new segment if the orignal stream is not exhausted
        if not data and not self.end_of_file:
            LOG.info("StreamReader: Finished serving data for segment %s. "
                     "Preparing for next.", self.segment)
            self._prepare_for_new_segment = True
            self.file_number += 1
            return b''

        # If we've reached the end of the file, just return empty
        if not data and self.end_of_file:
            return b''

        return data

    def seek(self, offset, whence=io.SEEK_SET):
        """Seek within the current segment's buffered data."""
        if not self._current_segment_buffer:
            raise io.UnsupportedOperation("StreamReader: No segment buffer "
                                          "available for seeking.")

        new_pos = self._current_segment_buffer.seek(offset, whence)
        self._buffer_read_offset = new_pos

        if new_pos > self._current_segment_length:
            raise IOError(f"StreamReader: Cannot seek beyond buffered data. "
                          f"Requested position: {new_pos}, Buffered data "
                          f"length: {self._current_segment_length}")

        return new_pos

    def tell(self):
        # Return the current position within the current segment's
        # buffered data.
        if not self._current_segment_buffer:
            return 0
        return self._buffer_read_offset

    @property
    def segment_checksum(self):
        """Returns the checksum of the *current* segment."""
        return self._current_segment_checksum.hexdigest()

    @property
    def segment_length(self):
        """Returns the length of the *current* segment."""
        return self._current_segment_length

    def release_buffer(self):
        """Manually release the current segment buffer."""
        if self._current_segment_buffer:
            self._current_segment_buffer.close()
            self._current_segment_buffer = None


class SwiftStorage(base.Storage):
    def __init__(self):
        self.client = _get_service_client(CONF.os_auth_url, CONF.os_token,
                                          CONF.os_tenant_id,
                                          region_name=CONF.os_region_name)

    def save(self, stream, metadata=None, container='database_backups'):
        """Persist data from the stream to swift.

        * Read data from stream, upload to swift
        * Update the new object metadata, stream provides method to get
          metadata.

        :returns the new object checkshum and swift full URL.
        """
        filename = stream.manifest
        LOG.info('Saving %(filename)s to %(container)s in swift.',
                 {'filename': filename, 'container': container})

        # Create the container if it doesn't already exist
        LOG.debug('Ensuring container %s', container)
        try:
            self.client.put_container(container)
        except swift_exc.ClientException as e:
            # 409 Conflict means container already exists
            if e.http_status != 409:
                LOG.error('Failed to create container %s: %s',
                          container, str(e))
                raise

        # Swift Checksum is the checksum of the concatenated segment checksums
        swift_checksum = hashlib.md5()
        # Wrap the output of the backup process to segment it for swift
        stream_reader = StreamReader(stream, container, filename,
                                     2 * (1024 ** 3))

        url = self.client.url
        # Full location where the backup manifest is stored
        location = "%s/%s/%s" % (url, container, filename)
        LOG.info('Uploading to %s', location)

        # Information about each segment upload job
        segment_results = []

        # Read from the stream and write to the container in swift
        while not stream_reader.end_of_file:
            segment_name = stream_reader.segment
            # This path includes the correct segment number
            path = stream_reader.segment_path

            LOG.info('Uploading segment %s.', segment_name)
            # self.client.put_object now receives a seekable stream_reader,
            # allowing swiftclient's internal retries to work on the current
            # segment.
            try:
                etag = self.client.put_object(container,
                                              segment_name,
                                              stream_reader)
            except swift_exc.ClientException as e:
                LOG.error('Swift client error uploading segment %s: %s',
                          segment_name, str(e))
                raise
            # After put_object returns, the segment_checksum and segment_length
            # properties of stream_reader refer to the segment that was just
            # uploaded.
            segment_md5 = stream_reader.segment_checksum
            current_segment_bytes = stream_reader.segment_length

            # Check each segment MD5 hash against swift etag
            if etag != segment_md5:
                msg = ('Failed to upload data segment to swift. ETAG: %(tag)s '
                       'Segment MD5: %(checksum)s.' %
                       {'tag': etag, 'checksum': segment_md5})
                raise Exception(msg)

            segment_results.append({
                'path': path,
                'etag': etag,
                'size_bytes': current_segment_bytes
            })

            swift_checksum.update(segment_md5.encode())

        # All segments uploaded.
        num_segments = len(segment_results)
        LOG.debug('File uploaded in %s segments.', num_segments)

        # An SLO will be generated if the backup was more than one segment in
        # length.
        large_object = num_segments > 1

        # Meta data is stored as headers
        if metadata is None:
            metadata = {}
        metadata.update(stream.get_metadata())
        headers = {}
        for key, value in metadata.items():
            headers[_set_attr(key)] = value

        LOG.info('Metadata headers: %s', headers)
        if large_object:
            manifest_data = json.dumps(segment_results)
            LOG.info('Creating the SLO manifest file, manifest content: %s',
                     manifest_data)
            # The etag returned from the manifest PUT is the checksum of the
            # manifest object (which is empty); this is not the checksum we
            # want.
            self.client.put_object(container,
                                   filename,
                                   manifest_data,
                                   headers=headers,
                                   query_string='multipart-manifest=put')

            # Validation checksum is the Swift Checksum
            final_swift_checksum = swift_checksum.hexdigest()
        else:
            LOG.info('Moving segment %(segment)s to %(filename)s.',
                     {'segment': stream_reader.first_segment,
                      'filename': filename})
            segment_result = segment_results[0]
            # Just rename it via a special put copy.
            headers['X-Copy-From'] = segment_result['path']
            self.client.put_object(container,
                                   filename, '',
                                   headers=headers)

            # Delete the old segment file that was copied
            LOG.info('Deleting the old segment file %s.',
                     stream_reader.first_segment)
            try:
                self.client.delete_object(container,
                                          stream_reader.first_segment)
            except swift_exc.ClientException as e:
                if e.http_status != 404:
                    raise

            final_swift_checksum = segment_result['etag']

        # Validate the object by comparing checksums
        resp = self.client.head_object(container, filename)
        # swift returns etag in double quotes
        # e.g. '"dc3b0827f276d8d78312992cc60c2c3f"'
        etag = resp['etag'].strip('"')

        # Raise an error and mark backup as failed
        if etag != final_swift_checksum:
            msg = ('Failed to upload data to swift. Manifest ETAG: %(tag)s '
                   'Swift MD5: %(checksum)s' %
                   {'tag': etag, 'checksum': final_swift_checksum})
            raise Exception(msg)

        return (final_swift_checksum, location)

    def _explodeLocation(self, location):
        storage_url = "/".join(location.split('/')[:-2])
        container = location.split('/')[-2]
        filename = location.split('/')[-1]
        return storage_url, container, filename

    def _verify_checksum(self, etag, checksum):
        etag_checksum = etag.strip('"')
        if etag_checksum != checksum:
            msg = ('Checksum validation failure, actual: %s, expected: %s' %
                   (etag_checksum, checksum))
            raise Exception(msg)

    def load(self, location, backup_checksum):
        """Get object from the location."""
        storage_url, container, filename = self._explodeLocation(location)

        headers, contents = self.client.get_object(container, filename,
                                                   resp_chunk_size=2 ** 16)

        if backup_checksum:
            self._verify_checksum(headers.get('etag', ''), backup_checksum)

        return contents

    def load_metadata(self, parent_location, parent_checksum):
        """Load metadata from swift."""
        if not parent_location:
            return {}

        _, container, filename = self._explodeLocation(parent_location)
        headers = self.client.head_object(container, filename)

        if parent_checksum:
            self._verify_checksum(headers.get('etag', ''), parent_checksum)

        _meta = {}
        for key, value in headers.items():
            if key.startswith('x-object-meta'):
                _meta[_get_attr(key)] = value

        return _meta

    def is_incremental_backup(self, location):
        """Check if the location is an incremental backup."""
        _, container, filename = self._explodeLocation(location)
        headers = self.client.head_object(container, filename)

        if 'x-object-meta-parent-location' in headers:
            return True

        return False

    def get_backup_lsn(self, location):
        """Get the backup LSN if exists."""
        _, container, filename = self._explodeLocation(location)
        headers = self.client.head_object(container, filename)
        return headers.get('x-object-meta-lsn')



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import re

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging

from backup.drivers import mysql_base

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class MariaBackup(mysql_base.MySQLBaseRunner):
    """Implementation of Backup and Restore using mariadb-backup."""
    restore_cmd = ('mbstream -x -C %(restore_location)s')
    prepare_cmd = 'mariadb-backup --prepare --target-dir=%(restore_location)s'

    def __init__(self, *args, **kwargs):
        super(MariaBackup, self).__init__(*args, **kwargs)
        self.backup_log = '/tmp/mariadb-backup.log'
        self._gzip = True

    @property
    def cmd(self):
        cmd = ('mariadb-backup --backup --stream=xbstream ' +
               self.user_and_pass)
        return cmd

    def check_restore_process(self):
        LOG.info('Checking return code of mbstream restore process.')
        return_code = self.process.returncode
        if return_code != 0:
            LOG.error('mbstream exited with %s', return_code)
            return False

        return True

    def post_restore(self):
        """Prepare after data restore."""
        LOG.info("Running prepare command: %s.", self.prepare_command)
        stdout, stderr = processutils.execute(*self.prepare_command.split())
        LOG.info("The command: %s : stdout: %s, stderr: %s",
                 self.prepare_command, stdout, stderr)
        LOG.info("Checking prepare log")
        if not stderr:
            msg = "Empty prepare log file"
            raise Exception(msg)
        last_line = stderr.splitlines()[-1].strip()
        if not re.search('completed OK!', last_line):
            msg = "Prepare did not complete successfully"
            raise Exception(msg)


class MariaBackupIncremental(MariaBackup):
    """Incremental backup and restore using mariadb-backup."""
    incremental_prep = ('mariadb-backup --prepare '
                        '--target-dir=%(restore_location)s '
                        '%(incremental_args)s')

    def __init__(self, *args, **kwargs):
        if not kwargs.get('lsn'):
            raise AttributeError('lsn attribute missing')
        self.parent_location = kwargs.pop('parent_location', '')
        self.parent_checksum = kwargs.pop('parent_checksum', '')

        super(MariaBackupIncremental, self).__init__(*args, **kwargs)

    @property
    def cmd(self):
        cmd = (
            'mariadb-backup --backup --stream=xbstream'
            ' --incremental-lsn=%(lsn)s ' +
            self.user_and_pass
        )
        LOG.info('cmd:{}'.format(cmd))
        return cmd

    def get_metadata(self):
        meta = super(MariaBackupIncremental, self).get_metadata()

        meta.update({
            'parent_location': self.parent_location,
            'parent_checksum': self.parent_checksum,
        })
        return meta

    def run_restore(self):
        """Run incremental restore."""
        LOG.info('Running incremental restore')
        self.incremental_restore(self.location, self.checksum)
        return self.restore_content_length



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

import docker

from trove.guestagent.utils import docker as docker_utils
from trove.tests.unittests import trove_testtools


class TestDockerUtils(trove_testtools.TestCase):
    def setUp(self):
        super().setUp()
        self.docker_client = mock.MagicMock()

    def test_create_network_with_network_exists(self):
        network_name = "test_network"
        network1 = mock.MagicMock(id="111")
        network1.name = "test_network"
        network2 = mock.MagicMock(id="222")
        network2.name = "test_network_2"
        self.docker_client.networks.list.return_value = [network1, network2]
        id = docker_utils.create_network(self.docker_client, network_name)
        self.assertEqual(id, "111")

    def test_create_network_ipv4_only(self):
        network_name = "test_network"
        eth1_data = json.dumps({"mac_address": "fa:16:3e:7c:9c:57",
                                "ipv4_address": "10.111.0.8",
                                "ipv4_cidr": "10.111.0.0/26",
                                "ipv4_gateway": "10.111.0.1",
                                "ipv4_host_routes": [{
                                    "destination": "10.10.0.0/16",
                                    "nexthop": "10.111.0.10"}]})
        net = mock.MagicMock(return_value=mock.MagicMock(id=111))
        self.docker_client.networks.create = net
        mo = mock.mock_open(read_data=eth1_data)
        with mock.patch.object(docker_utils, 'open', mo):
            id = docker_utils.create_network(self.docker_client, network_name)
        self.assertEqual(id, 111)
        net.assert_called_once()
        kwargs = net.call_args.kwargs
        self.assertEqual(kwargs.get("name"), "test_network")
        self.assertEqual(kwargs.get("driver"), "docker-hostnic")
        self.assertEqual(len(kwargs.get("ipam").get("Config")), 1)
        self.assertEqual(kwargs["ipam"]["Config"][0]["Gateway"],
                         "10.111.0.1")
        self.assertEqual(kwargs["enable_ipv6"], False)
        self.assertEqual(kwargs["options"]["hostnic_mac"],
                         "fa:16:3e:7c:9c:57")

    def test_create_network_ipv6_only(self):
        network_name = "test_network"
        eth1_data = json.dumps({"mac_address": "fa:16:3e:7c:9c:58",
                                "ipv6_address":
                                    "fda3:96d9:23e:0:f816:3eff:fe7c:9c57",
                                "ipv6_cidr": "fda3:96d9:23e::/64",
                                "ipv6_gateway": "fda3:96d9:23e::1"})
        net = mock.MagicMock(return_value=mock.MagicMock(id=222))
        self.docker_client.networks.create = net
        mo = mock.mock_open(read_data=eth1_data)
        with mock.patch.object(docker_utils, 'open', mo):
            id = docker_utils.create_network(self.docker_client, network_name)
        self.assertEqual(id, 222)
        net.assert_called_once()
        kwargs = net.call_args.kwargs
        self.assertEqual(kwargs.get("name"), "test_network")
        self.assertEqual(kwargs.get("driver"), "docker-hostnic")
        self.assertEqual(len(kwargs.get("ipam").get("Config")), 1)
        self.assertEqual(kwargs["ipam"]["Config"][0]["Gateway"],
                         "fda3:96d9:23e::1")
        self.assertEqual(kwargs["enable_ipv6"], True)
        self.assertEqual(kwargs["options"]["hostnic_mac"], "fa:16:3e:7c:9c:58")

    def test_create_network_dual_stack(self):
        network_name = "test_network"
        eth1_data = json.dumps({"mac_address": "fa:16:3e:7c:9c:59",
                                "ipv4_address": "10.111.0.8",
                                "ipv4_cidr": "10.111.0.0/26",
                                "ipv4_gateway": "10.111.0.1",
                                "ipv4_host_routes": [{
                                    "destination": "10.10.0.0/16",
                                    "nexthop": "10.111.0.10"}],
                                "ipv6_address":
                                    "fda3:96d9:23e:0:f816:3eff:fe7c:9c57",
                                "ipv6_cidr": "fda3:96d9:23e::/64",
                                "ipv6_gateway": "fda3:96d9:23e::1"})
        net = mock.MagicMock(return_value=mock.MagicMock(id=333))
        self.docker_client.networks.create = net
        mo = mock.mock_open(read_data=eth1_data)
        with mock.patch.object(docker_utils, 'open', mo):
            id = docker_utils.create_network(self.docker_client, network_name)
        self.assertEqual(id, 333)
        net.assert_called_once()
        kwargs = net.call_args.kwargs
        self.assertEqual(kwargs["name"], "test_network")
        self.assertEqual(kwargs["driver"], "docker-hostnic")
        self.assertEqual(len(kwargs["ipam"]["Config"]), 2)
        self.assertEqual(kwargs["enable_ipv6"], True)
        self.assertEqual(kwargs["options"]["hostnic_mac"],
                         "fa:16:3e:7c:9c:59")

    @mock.patch("docker.APIClient")
    def test__create_container_with_low_level_api(self, mock_client):
        eth1_data = json.dumps({
            "mac_address": "fa:16:3e:7c:9c:57",
            "ipv4_address": "10.111.0.8",
            "ipv4_cidr": "10.111.0.0/26",
            "ipv4_gateway": "10.111.0.1",
            "ipv4_host_routes": [{"destination": "10.10.0.0/16",
                                  "nexthop": "10.111.0.10"}]})

        mo = mock.mock_open(read_data=eth1_data)
        param = dict(name="test",
                     restart_policy={"Name": "always"},
                     privileged=False,
                     detach=True,
                     volumes={},
                     ports={},
                     user="test_user",
                     environment={},
                     command="sleep inf")
        with mock.patch.object(docker_utils, 'open', mo):
            docker_utils._create_container_with_low_level_api(
                "busybox", param)
        mock_client().create_host_config.assert_called_once()
        mock_client().create_networking_config.assert_called_once()
        mock_client().pull.assert_called_once()
        mock_client().create_container.assert_called_once()
        mock_client().start.assert_called_once()

    @mock.patch("docker.DockerClient")
    def test_get_health_status(self, mock_client):
        mock_container = mock.MagicMock()
        mock_container.health = "healthy"
        mock_client.containers.get.return_value = mock_container

        health_status = docker_utils.get_container_health(mock_client,
                                                          "test_container")
        self.assertEqual(health_status, "healthy")
        mock_client.containers.get.assert_called_once_with("test_container")

    @mock.patch("docker.DockerClient")
    def test_get_health_status_not_found(self, mock_client):
        mock_client.containers.get.side_effect = docker.errors.NotFound(
            "Container not found")
        health_status = docker_utils.get_container_health(mock_client,
                                                          "test_container")
        self.assertEqual(health_status, "not running")
        mock_client.containers.get.assert_called_once_with("test_container")



# Copyright 2023 Yovole
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import jsonschema
import netaddr
import os
import sys
import traceback

import flask
from flask import Flask
import gunicorn.app.base
from oslo_log import log as logging
from pyroute2 import IPRoute
from werkzeug import exceptions as w_exceptions

from trove.common import constants
from trove.common import schemata

LOG = logging.getLogger(__name__)


class hostnic_config(object):
    """this class records network id and its host nic"""
    CONFIG_FILE = "/etc/docker/hostnic.json"

    def __init__(self) -> None:
        if not os.path.exists(self.CONFIG_FILE):
            with open(self.CONFIG_FILE, 'w+') as f:
                f.write(json.dumps({}))

    def get_data(self) -> dict:
        with open(self.CONFIG_FILE, 'r') as cfg:
            data = json.loads(cfg.read())
        return data

    def write_config(self, key: str, value: str):
        data = self.get_data()
        data[key] = value
        with open(self.CONFIG_FILE, 'w+') as cfg:
            cfg.write(json.dumps(data))

    def get_config(self, key: str):
        data = self.get_data()
        return data.get(key, "")

    def delete_config(self, key: str):
        data = self.get_data()
        if not data.get(key):
            return
        data.pop(key)
        with open(self.CONFIG_FILE, 'w+') as cfg:
            cfg.write(json.dumps(data))


driver_config = hostnic_config()


def make_json_app(import_name, **kwargs):
    """Creates a JSON-oriented Flask app.

    All error responses that you don't specifically manage yourself will have
    application/json content type, and will contain JSON that follows the
    libnetwork remote driver protocol.


    { "Err": "405: Method Not Allowed" }


    See:
      - https://github.com/docker/libnetwork/blob/3c8e06bc0580a2a1b2440fe0792fbfcd43a9feca/docs/remote.md#errors  # noqa
    """

    app = Flask(import_name)

    @app.errorhandler(jsonschema.ValidationError)
    def make_json_error(ex):
        LOG.error("Unexpected error happened: %s", ex)
        traceback.print_exc(file=sys.stderr)
        response = flask.jsonify({"Err": str(ex)})
        response.status_code = w_exceptions.InternalServerError.code
        if isinstance(ex, w_exceptions.HTTPException):
            response.status_code = ex.code
        elif isinstance(ex, jsonschema.ValidationError):
            response.status_code = w_exceptions.BadRequest.code
        content_type = 'application/vnd.docker.plugins.v1+json; charset=utf-8'
        response.headers['Content-Type'] = content_type
        return response

    for code in w_exceptions.default_exceptions:
        app.register_error_handler(code, make_json_error)

    return app


app = make_json_app(__name__)


@app.route('/Plugin.Activate', methods=['POST', 'GET'])
def plugin_activate():
    """Returns the list of the implemented drivers.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#handshake  # noqa
    """
    LOG.debug("Received /Plugin.Activate")
    return flask.jsonify(schemata.SCHEMA['PLUGIN_ACTIVATE'])


@app.route('/NetworkDriver.GetCapabilities', methods=['POST'])
def plugin_scope():
    """Returns the capability as the remote network driver.

    This function returns the capability of the remote network driver, which is
    ``global`` or ``local`` and defaults to ``local``. With ``global``
    capability, the network information is shared among multipe Docker daemons
    if the distributed store is appropriately configured.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#set-capability  # noqa
    """
    LOG.debug("Received /NetworkDriver.GetCapabilities")
    capabilities = {'Scope': 'local'}
    return flask.jsonify(capabilities)


@app.route('/NetworkDriver.DiscoverNew', methods=['POST'])
def network_driver_discover_new():
    """The callback function for the DiscoverNew notification.

    The DiscoverNew notification includes the type of the
    resource that has been newly discovered and possibly other
    information associated with the resource.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#discovernew-notification  # noqa
    """
    LOG.debug("Received /NetworkDriver.DiscoverNew")
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.DiscoverDelete', methods=['POST'])
def network_driver_discover_delete():
    """The callback function for the DiscoverDelete notification.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#discoverdelete-notification  # noqa
    """
    LOG.debug("Received /NetworkDriver.DiscoverDelete")
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.CreateNetwork', methods=['POST'])
def network_driver_create_network():
    """Creates a new  Network which name is the given NetworkID.
    example:
    docker network create --driver docker-hostnic --gateway 192.168.1.1  --subnet 192.168.1.0/24 -o hostnic_mac=52:54:00:e1:d9:ef  test_network
    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#create-network  # noqa
    """
    json_data = flask.request.get_json(force=True)
    jsonschema.validate(json_data, schemata.NETWORK_CREATE_SCHEMA)
    hostnic_mac = \
        json_data['Options']['com.docker.network.generic']['hostnic_mac']
    if driver_config.get_config(json_data['NetworkID']):
        return flask.jsonify("network already has a host nic")
    gw = json_data.get("IPv4Data")[0].get("Gateway", '')
    netinfo = {"mac_address": hostnic_mac}
    if gw:
        ip = netaddr.IPNetwork(gw)
        netinfo["gateway"] = str(ip.ip)
    driver_config.write_config(json_data['NetworkID'], netinfo)
    LOG.debug("Received JSON data %s for /NetworkDriver.Create", json_data)
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.DeleteNetwork', methods=['POST'])
def network_driver_delete_network():
    # Just remove the network from the config file.
    json_data = flask.request.get_json(force=True)
    driver_config.delete_config(json_data['NetworkID'])
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.Join', methods=['POST'])
def network_driver_join():
    json_data = flask.request.get_json(force=True)
    jsonschema.validate(json_data, schemata.NETWORK_JOIN_SCHEMA)
    netid = json_data['NetworkID']
    hostnic_mac = driver_config.get_config(netid).get('mac_address')
    ipr = IPRoute()
    ifaces = ipr.get_links(address=hostnic_mac)
    ifname = ifaces[0].get_attr('IFLA_IFNAME')
    with open(constants.ETH1_CONFIG_PATH) as fd:
        eth1_config = json.load(fd)
    join_response = {
        "InterfaceName": {
            "SrcName": ifname,
            "DstPrefix": "eth"},
    }
    if eth1_config.get("ipv4_gateway"):
        join_response["Gateway"] = eth1_config.get("ipv4_gateway")
    if eth1_config.get("ipv6_gateway"):
        join_response["GatewayIPv6"] = eth1_config.get("ipv6_gateway")
    if eth1_config.get("ipv4_host_routes"):
        join_response["StaticRoutes"] = list()
        for route in eth1_config.get("ipv4_host_routes"):
            join_response["StaticRoutes"].append(
                {"Destination": route["destination"],
                 "NextHop": route["nexthop"]})
    return flask.jsonify(join_response)


@app.route('/NetworkDriver.Leave', methods=['POST'])
def network_driver_leave():
    """Unbinds a hostnic from a sandbox.

    This function takes the following JSON data and delete the veth pair
    corresponding to the given info. ::

        {
            "NetworkID": string,
            "EndpointID": string
        }
    we don't need to remove the port from the sandbox explicitly,
    once the sandbox get deleted, the hostnic comes to default
    netns automatically.
    """
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.DeleteEndpoint', methods=['POST'])
def network_driver_delete_endpoint():
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.CreateEndpoint', methods=['POST'])
def network_driver_create_endpoint():
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.EndpointOperInfo', methods=['POST'])
def network_driver_endpoint_operational_info():
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.ProgramExternalConnectivity', methods=['POST'])
def network_driver_program_external_connectivity():
    """provide external connectivity for the given container."""
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.RevokeExternalConnectivity', methods=['POST'])
def network_driver_revoke_external_connectivity():
    """Removes external connectivity for a given container.

    Performs the necessary programming to remove the external connectivity
    of a container

    See the following link for more details about the spec:
      https://github.com/docker/libnetwork/blob/master/driverapi/driverapi.go
    """
    return flask.jsonify(schemata.SCHEMA['SUCCESS'])


class StandaloneApplication(gunicorn.app.base.BaseApplication):

    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def main():
    options = {
        'bind': "unix:/run/docker/docker-hostnic.sock",
        'workers': 1,
    }
    StandaloneApplication(app, options).run()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

import unittest
from unittest.mock import MagicMock

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
CONF.backup_encryption_key = None
CONF.backup_id = "backup_unittest"

driver_mapping = {
    'innobackupex': 'backup.drivers.innobackupex.InnoBackupEx',
    'innobackupex_inc': 'backup.drivers.innobackupex.InnoBackupExIncremental',
    'mariabackup': 'backup.drivers.mariabackup.MariaBackup',
    'mariabackup_inc': 'backup.drivers.mariabackup.MariaBackupIncremental',
    'pg_basebackup': 'backup.drivers.postgres.PgBasebackup',
    'pg_basebackup_inc': 'backup.drivers.postgres.PgBasebackupIncremental',
    'xtrabackup': 'backup.drivers.xtrabackup.XtraBackup',
    'xtrabackup_inc': 'backup.drivers.xtrabackup.XtraBackupIncremental'
}


class TestPgBasebackup(unittest.TestCase):
    def setUp(self):
        self.runner_cls = importutils.import_class(
            driver_mapping['pg_basebackup'])
        self.params = {
            'wal_archive_dir': './',
            'filename': '000000010000000000000006.00000168.backup'
        }

        # assertions
        self.assertIsNotNone(self.runner_cls)

    def tearDown(self):
        if os.path.exists(self.params.get('filename')):
            os.remove(self.params.get('filename'))

    def _create_test_data(self):
        with open(self.params.get('filename'), 'w') as file:
            file.write("START WAL LOCATION: -1/3000028 "
                       "(file 000000010000000000000003)\n")
            file.write("STOP WAL LOCATION: 0/3000028 "
                       "(file 000000010000000000000003)\n")
            file.write("CHECKPOINT LOCATION: 0/3000098\n")
            file.write("BACKUP METHOD: streamed\n")
            file.write("BACKUP FROM: master\n")
            file.write("START TIME: 2023-05-01 06:53:41 UTC\n")
            file.write("LABEL: 3070d460-1e67-4fbd-92ca-97c1d0101077\n")
            file.write("START TIMELINE: 1\n")

    def test_instance(self):
        '''Check instance'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        self.assertIsNotNone(runner)

    def test_cmd(self):
        '''Check cmd property'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        self.assertEqual(runner.cmd,
                         "pg_basebackup -U postgres -Ft -z "
                         "--wal-method=fetch --label={} "
                         "--pgdata=-".format(self.params.get('filename')))

    def test_manifest(self):
        '''Check manifest'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        self.assertEqual(runner.manifest,
                         "{}.tar.gz".format(self.params.get('filename')))

    def test_is_read_only(self):
        '''Check is_read_only'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        runner._is_read_only = True
        self.assertEqual(runner.is_read_only, True)

    def test_get_wal_files(self):
        '''Check get_wal_file'''
        # prepare the test
        runner = self.runner_cls(**self.params)
        recent_backup_file = "000000010000000000000006.00000168.backup"
        last_wal = "000000010000000000000007"
        self._create_test_data()

        runner.get_backup_files = MagicMock(
            return_value=[recent_backup_file])
        with open(last_wal, "w") as file:
            file.write("test")

        # call the method
        ret = runner.get_wal_files()

        # assertions
        self.assertEqual(ret, [last_wal])

        if os.path.exists(last_wal):
            os.remove(last_wal)

    def test_get_backup_files(self):
        '''Check get_backup_file'''
        # prepare the test
        runner = self.runner_cls(**self.params)
        recent_backup_file = "000000010000000000000006.00000168.backup"
        runner.get_backup_files = MagicMock(
            return_value=[recent_backup_file])

        # call the method
        ret = runner.get_backup_files()

        # assertions
        self.assertEqual(ret, [recent_backup_file])

    def test_get_backup_metadata(self):
        '''Check get_backup_metadata'''
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner.label = self.params.get('filename')
        self._create_test_data()

        # call the method
        backup_metadata = runner.get_backup_metadata(
            self.params.get('filename')
        )

        # assertions
        self.assertEqual(backup_metadata['start-segment'], '-1/3000028')
        self.assertEqual(
            backup_metadata['start-wal-file'], '000000010000000000000003'
        )
        self.assertEqual(backup_metadata['stop-segment'], '0/3000028')
        self.assertEqual(
            backup_metadata['stop-wal-file'], '000000010000000000000003')
        self.assertEqual(
            backup_metadata['checkpoint-location'], '0/3000098'
        )
        self.assertEqual(
            backup_metadata['label'], '3070d460-1e67-4fbd-92ca-97c1d0101077'
        )

    def test_get_metadata(self):
        '''Check get_metadata'''
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner.get_metadata = MagicMock(
            return_value={'start-segment': '0/3000028'}
        )

        # call the method
        metadata = runner.get_metadata()

        # assertions
        self.assertEqual(metadata['start-segment'], '0/3000028')

    def test_context(self):
        '''Check context methods'''
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner._is_read_only = True
        runner.pre_backup = MagicMock()
        runner._run = MagicMock()
        runner.post_backup = MagicMock()

        # call the method
        with runner:
            pass

        # assertions
        runner.pre_backup.assert_called_once_with()
        runner._run.assert_called_once_with()
        runner.post_backup.assert_called_once_with()

    def test_check_process(self):
        '''Check check_process'''
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner._is_read_only = True
        runner.start_segment = True
        runner.start_wal_file = True
        runner.stop_segment = True
        runner.stop_wal_file = True
        runner.label = True

        # call the method
        ret = runner.check_process()

        # assertions
        self.assertTrue(ret)

    def test_check_restore_process(self):
        '''Check check_restore_process'''
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner._is_read_only = True
        runner.start_segment = True
        runner.start_wal_file = True
        runner.stop_segment = True
        runner.stop_wal_file = True
        runner.label = True

        # call the method
        ret = runner.check_process()

        # assertions
        self.assertTrue(ret)


class TestPgBasebackupIncremental(unittest.TestCase):
    def setUp(self):
        self.runner_cls = importutils.import_class(
            driver_mapping['pg_basebackup_inc'])
        self.params = {
            'wal_archive_dir': './',
            'filename': '000000010000000000000006.00000168.backup',
            'parent_location': 'http://example.com/example.tar.gz',
            'parent_checksum': '63e696c5eb85550fed0a7a1a6411eb7d'
        }
        self.metadata = {
            'start-segment': '0/3000028',
            'start-wal-file': '000000010000000000000003',
            'stop-segment': '0/3000028',
            'stop-wal-file': '000000010000000000000003',
            'checkpoint-location': '0/3000098',
            'label': '000000010000000000000006.00000168.backup',
            'parent_location': self.params.get('parent_location'),
            'parent_checksum': self.params.get('parent_checksum'),
        }

    def tearDown(self):
        if os.path.exists(self.params.get('filename')):
            os.remove(self.params.get('filename'))

    def test_instance(self):
        '''Check instance'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        self.assertIsNotNone(runner)

    def test_pre_backup(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner.pre_backup = MagicMock(return_value=None)

        # call the method
        runner.pre_backup()

        # assertions
        runner.pre_backup.assert_called_once_with()

    def test_cmd(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        wal_file_list = [
            '000000010000000000000005',
            '000000010000000000000003',
            '000000010000000000000004'
        ]
        wal_archive_dir = self.params.get('wal_archive_dir')
        cmd = (f'tar -czf - -C {wal_archive_dir} '
               f'{" ".join(wal_file_list)}')

        runner.get_wal_files = MagicMock(return_value=wal_file_list)

        # call the method
        ret = runner._cmd()

        # assertions
        self.assertEqual(ret, cmd)

    def test_get_metadata(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner.get_metadata = MagicMock(return_value=self.metadata)

        # call the method
        ret = runner.get_metadata()

        # assertions
        self.assertEqual(ret, self.metadata)

    def test_incremental_restore_cmd(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        cmd = 'tar xzf - -C /var/lib/postgresql/data/pgdata'

        # call the method
        ret = runner.incremental_restore_cmd()

        # assertions
        self.assertEqual(ret, cmd)

    def test_incremental_restore(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        wal_file_list = [
            '000000010000000000000005',
            '000000010000000000000003',
            '000000010000000000000004'
        ]
        runner.get_wal_files = MagicMock(return_value=wal_file_list)
        metadata = {
            'parent_location': 'https://example.com/',
            'parent_checksum': 'cc39f022c5d10f38e963062ca40c95bd',
        }
        runner.storage = MagicMock(return_value=metadata)
        command = "testcommand"
        length = 10
        runner.incremental_restore = MagicMock(return_value=length)
        runner.incremental_restore_cmd = MagicMock(return_value=command)
        runner.unpack = MagicMock(return_value=length)

        # call the method
        ret = runner.incremental_restore({
            'location': metadata['parent_location'],
            'checksum': metadata['parent_checksum']
        })

        # assertions
        self.assertEqual(ret, length)

    def test_run_restore(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        length = 10
        runner.incremental_restore = MagicMock(return_value=length)
        runner.restore_content_length = length

        # call the method
        ret = runner.run_restore()

        # assertions
        self.assertEqual(ret, length)


if __name__ == '__main__':
    unittest.main()



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import MetaData
from sqlalchemy import orm
from sqlalchemy.orm import exc as orm_exc
from sqlalchemy.orm import registry
from sqlalchemy import Table

mapper_registry = registry()


def map(engine, models):
    meta = MetaData()
    meta.bind = engine
    if mapping_exists(models['instances']):
        return

    mapper_registry.map_imperatively(models['instances'],
                                     Table('instances', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['instance_faults'],
                                     Table('instance_faults', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['root_enabled_history'],
                                     Table('root_enabled_history', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['datastores'],
                                     Table('datastores', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['datastore_versions'],
                                     Table('datastore_versions', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['datastore_version_metadata'],
                                     Table('datastore_version_metadata',
                                           meta, autoload_with=engine))
    mapper_registry.map_imperatively(models['capabilities'],
                                     Table('capabilities', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['capability_overrides'],
                                     Table('capability_overrides', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['service_statuses'],
                                     Table('service_statuses', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['dns_records'],
                                     Table('dns_records', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['agent_heartbeats'],
                                     Table('agent_heartbeats', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['quotas'],
                                     Table('quotas', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['quota_usages'],
                                     Table('quota_usages', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['reservations'],
                                     Table('reservations', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['backups'],
                                     Table('backups', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['backup_strategy'],
                                     Table('backup_strategy', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['security_groups'],
                                     Table('security_groups', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['security_group_rules'],
                                     Table('security_group_rules', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(
        models['security_group_instance_associations'],
        Table('security_group_instance_associations', meta,
              autoload_with=engine))
    mapper_registry.map_imperatively(models['configurations'],
                                     Table('configurations', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['configuration_parameters'],
                                     Table('configuration_parameters',
                                           meta, autoload_with=engine))
    mapper_registry.map_imperatively(models['conductor_lastseen'],
                                     Table('conductor_lastseen', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['clusters'],
                                     Table('clusters', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(
        models['datastore_configuration_parameters'],
        Table('datastore_configuration_parameters', meta,
              autoload_with=engine))
    mapper_registry.map_imperatively(models['modules'],
                                     Table('modules', meta,
                                           autoload_with=engine))
    mapper_registry.map_imperatively(models['instance_modules'],
                                     Table('instance_modules', meta,
                                           autoload_with=engine))


def mapping_exists(model):
    try:
        orm.class_mapper(model)
        return True
    except orm_exc.UnmappedClassError:
        return False



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from trove.guestagent.datastore.postgres import service
from trove.tests.unittests import trove_testtools

statement = (
    "SELECT usename FROM pg_catalog.pg_user WHERE usesuper = true"
)
username = "postgres"
port = 5432
result = [(1, "one"), (2, 'two'), (3, 'three')]


class TestPostgresConnection(trove_testtools.TestCase):
    def setUp(self):
        super(TestPostgresConnection, self).setUp()

    # execute is expected to returns nothing
    @mock.patch(
        'trove.guestagent.datastore.postgres.service.PostgresConnection')
    def test_execute(self, mock_postgres_connection):
        postgres_connection = mock.MagicMock()
        postgres_connection.execute = mock.MagicMock(return_value=None)
        mock_postgres_connection.return_value = postgres_connection

        # assertion here
        connection = service.PostgresConnection(username, port=port)
        self.assertIsNone(connection.execute(statement),
                          'postgres_connection_execute does not returns None')

    # query is expected to returns result
    @mock.patch(
        'trove.guestagent.datastore.postgres.service.PostgresConnection')
    def test_query(self, mock_postgres_connection):
        postgres_connection = mock.MagicMock()
        postgres_connection.query = mock.MagicMock(return_value=result)
        mock_postgres_connection.return_value = postgres_connection

        # assertion here
        connection = service.PostgresConnection(username, port=port)
        self.assertEqual(result, connection.query(statement),
                         'postgres_connection_query does not returns expected')



# Copyright 2014 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
import semantic_version

from trove.common import cfg
from trove.guestagent.strategies.replication import mysql_base

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class MysqlGTIDReplication(mysql_base.MysqlReplicationBase):
    """MySql Replication coordinated by GTIDs."""

    def connect_to_master(self, service, master_info):
        cur_ver = semantic_version.Version.coerce(CONF.datastore_version)
        mysql_84 = semantic_version.Version('8.4.0')
        if 'dataset' in master_info:
            # pull the last executed GTID from the master via
            # the xtrabackup metadata file. If that value is
            # provided we need to set the gtid_purged variable
            # before executing the CHANGE MASTER TO command
            if master_info.get('dataset', {}).get('log_executed_gtid_set'):
                last_gtid = master_info['dataset']['log_executed_gtid_set']
            else:
                last_gtid = self.read_last_master_gtid(service)
            LOG.info("last_gtid value is %s", last_gtid)
            if '-' in last_gtid:
                # Also, FLUSH PRIVILEGES will restore gtid_executed.
                if cur_ver >= mysql_84:
                    service.execute_sql('RESET BINARY LOGS AND GTIDS')
                else:
                    # for mysql 8.0
                    service.execute_sql('RESET MASTER')
                set_gtid_cmd = "SET GLOBAL gtid_purged='%s'" % last_gtid
                service.execute_sql(set_gtid_cmd)

        replica_conf = master_info['replica_conf']
        LOG.info(
            "Configure the slave, master: %s:%s, replication user: %s",
            master_info['master']['host'],
            master_info['master']['port'],
            replica_conf['replication_user']['name']
        )
        if cur_ver >= mysql_84:
            change_master_cmd = (
                "CHANGE REPLICATION SOURCE TO "
                "SOURCE_HOST='%(host)s', "
                "SOURCE_PORT=%(port)s, "
                "SOURCE_USER='%(user)s', "
                "SOURCE_PASSWORD='%(password)s', "
                "SOURCE_AUTO_POSITION=1, "
                "GET_SOURCE_PUBLIC_KEY=1, "
                "SOURCE_CONNECT_RETRY=15" %
                {
                    'host': master_info['master']['host'],
                    'port': master_info['master']['port'],
                    'user': replica_conf['replication_user']['name'],
                    'password': replica_conf['replication_user']['password']
                })
        else:
            change_master_cmd = (
                "CHANGE MASTER TO "
                "MASTER_HOST='%(host)s', "
                "MASTER_PORT=%(port)s, "
                "MASTER_USER='%(user)s', "
                "MASTER_PASSWORD='%(password)s', "
                "MASTER_AUTO_POSITION=1, "
                "MASTER_CONNECT_RETRY=15" %
                {
                    'host': master_info['master']['host'],
                    'port': master_info['master']['port'],
                    'user': replica_conf['replication_user']['name'],
                    'password': replica_conf['replication_user']['password']
                })

        service.execute_sql(change_master_cmd)
        service.start_slave()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from unittest.mock import Mock, MagicMock

from trove.common import exception
from trove.db.sqlalchemy import api


class TestDbSqlalchemyApi(unittest.TestCase):

    def test_db_sync_alembic(self):
        api._configure_alembic = MagicMock(return_value=True)
        api._get_alembic_revision = MagicMock(return_value='head')
        api.alembic_command.upgrade = Mock()
        api.db_sync({})
        self.assertTrue(api.alembic_command.upgrade.called)

    def test_db_sync_sqlalchemy_migrate(self):
        api._configure_alembic = MagicMock(return_value=False)
        with self.assertRaises(exception.BadRequest) as ex:
            api.db_sync({})
            self.assertTrue(ex.msg,
                            'sqlalchemy-migrate is no longer supported')

    def test_db_upgrade_alembic(self):
        api._configure_alembic = MagicMock(return_value=True)
        api.alembic_command.upgrade = Mock()
        api.db_upgrade({})
        self.assertTrue(api.alembic_command.upgrade.called)

    def test_db_upgrade_sqlalchemy_migrate(self):
        api._configure_alembic = MagicMock(return_value=False)
        with self.assertRaises(exception.BadRequest) as ex:
            api.db_upgrade({})
            self.assertTrue(ex.msg,
                            'sqlalchemy-migrate is no longer supported')



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import semantic_version

from sqlalchemy.sql.expression import text

from oslo_log import log as logging
from oslo_utils.excutils import save_and_reraise_exception
from trove.common import cfg
from trove.common import constants
from trove.guestagent.datastore.mysql_common import service
from trove.guestagent.utils import docker as docker_util
from trove.guestagent.utils import mysql as mysql_util


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class MySqlApp(service.BaseMySqlApp):

    HEALTHCHECK = {
        "test": ["CMD", "mysqladmin", "ping", "-h",
                 "127.0.0.1", "-u", "root",
                 "--password=$MYSQL_ROOT_PASSWORD"],
        "start_period": 10 * 1000000000,  # 10 seconds in nanoseconds
        "interval": 10 * 1000000000,
        "timeout": 5 * 1000000000,
        "retries": 3
    }

    def _is_mysql84(self):
        mysql_84 = semantic_version.Version('8.4.0')
        cur_ver = semantic_version.Version.coerce(CONF.datastore_version)
        if cur_ver >= mysql_84:
            return True
        else:
            return False

    def __init__(self, status, docker_client):
        super(MySqlApp, self).__init__(status, docker_client)

    def _get_gtid_executed(self):
        with mysql_util.SqlClient(self.get_engine()) as client:
            return client.execute(
                text('SELECT @@global.gtid_executed')).first()[0]

    def _get_slave_status(self):
        with mysql_util.SqlClient(self.get_engine()) as client:
            if self._is_mysql84():
                return client.execute(text('SHOW REPLICA STATUS')).first()
            else:
                return client.execute(text('SHOW SLAVE STATUS')).first()

    def _get_master_UUID(self):
        slave_status = self._get_slave_status()
        if self._is_mysql84():
            return slave_status and slave_status._mapping['Source_UUID'] \
                or None
        else:
            return slave_status and slave_status._mapping['Master_UUID'] \
                or None

    def get_latest_txn_id(self):
        return self._get_gtid_executed()

    def get_last_txn(self):
        master_UUID = self._get_master_UUID()
        last_txn_id = '0'
        gtid_executed = self._get_gtid_executed()
        for gtid_set in gtid_executed.split(','):
            uuid_set = gtid_set.split(':')
            if str(uuid_set[0]) == str(master_UUID):
                last_txn_id = uuid_set[-1].split('-')[-1]
                break
        return master_UUID, int(last_txn_id)

    def wait_for_txn(self, txn):
        if self._is_mysql84():
            _sql = "SELECT WAIT_FOR_EXECUTED_GTID_SET('%s')" % txn
        else:
            _sql = "SELECT WAIT_UNTIL_SQL_THREAD_AFTER_GTIDS('%s')" % txn
        with mysql_util.SqlClient(self.get_engine()) as client:
            client.execute(text(_sql))

    def stop_master(self):
        LOG.info("Stopping replication master.")
        if not self._is_mysql84():
            return super().stop_master()
        with mysql_util.SqlClient(self.get_engine()) as client:
            client.execute(text("RESET BINARY LOGS AND GTIDS"))

    def get_backup_image(self):
        """Get the actual container image based on datastore version.

        For example, this method converts openstacktrove/db-backup-mysql:1.0.0
        to openstacktrove/db-backup-mysql5.7:1.0.0

        **deprecated**: this function is for backward compatibility.
        """
        image = cfg.get_configuration_property('backup_docker_image')
        if not self._image_has_tag(image):
            return super().get_backup_image()
        else:
            name, tag = image.rsplit(':', 1)
            # Get minor version
            cur_ver = semantic_version.Version.coerce(CONF.datastore_version)
            minor_ver = f"{cur_ver.major}.{cur_ver.minor}"
            return f"{name}{minor_ver}:{tag}"

    def get_backup_strategy(self):
        """Get backup strategy.

        innobackupex was removed in Percona XtraBackup 8.0, use xtrabackup
        instead.
        """
        return 'xtrabackup'

    def reset_data_for_restore_snapshot(self, data_dir):
        """This function try remove replica status in database"""
        # '--skip-replica-start' was introduced in mysql 8.0.26 and the
        # '--skip-slave-start' not be removed yet for mysql 8.0.x
        if self._is_mysql84:
            command = "mysqld --skip-replica-start=ON --datadir=%s" % data_dir
        else:
            command = "mysqld --skip-slave-start=ON --datadir=%s" % data_dir

        extra_volumes = {
            "/etc/mysql": {"bind": "/etc/mysql", "mode": "rw"},
            constants.MYSQL_HOST_SOCKET_PATH: {
                "bind": "/var/run/mysqld", "mode": "rw"},
            data_dir: {"bind": data_dir, "mode": "rw"},
        }

        try:
            self.start_db(ds_version=CONF.datastore_version, command=command,
                          extra_volumes=extra_volumes)
            self.stop_slave(for_failover=False)
        except Exception as err:
            with save_and_reraise_exception():
                LOG.error("Failed to remove slave status: %s", str(err))
        finally:
            try:
                LOG.debug(
                    'The init container log: %s',
                    docker_util.get_container_logs(self.docker_client))
                docker_util.remove_container(self.docker_client)
            except Exception as err:
                LOG.error('Failed to remove container. error: %s', str(err))

    def start_slave(self):
        LOG.info("Starting slave replication.")
        if not self._is_mysql84():
            return super().start_slave()
        with mysql_util.SqlClient(self.get_engine()) as client:
            client.execute(text('START REPLICA'))
            self.wait_for_slave_status("ON", client, 180)

    def stop_slave(self, for_failover):
        LOG.info("Stopping slave replication.")
        if not self._is_mysql84():
            return super().stop_slave(for_failover)
        replication_user = None
        with mysql_util.SqlClient(self.get_engine()) as client:
            result = client.execute(
                text('SHOW REPLICA STATUS')).mappings().first()
            if result:
                replication_user = result['Source_User']
                client.execute(text('STOP REPLICA'))
                client.execute(text('RESET REPLICA ALL'))
            self.wait_for_slave_status('OFF', client, 180)
            if not for_failover and replication_user:
                client.execute(
                    text('DROP USER IF EXISTS ' + replication_user))
        return {
            'replication_user': replication_user
        }


class MySqlRootAccess(service.BaseMySqlRootAccess):
    def __init__(self, app):
        super(MySqlRootAccess, self).__init__(app)


class MySqlAdmin(service.BaseMySqlAdmin):
    def __init__(self, app):
        root_access = MySqlRootAccess(app)
        super(MySqlAdmin, self).__init__(root_access, app)



# Copyright 2021 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

BACKUP_TYPE_FULL = 'full'
BACKUP_TYPE_INC = 'incremental'
ETH1_CONFIG_PATH = "/etc/trove/eth1.json"
DOCKER_NETWORK_NAME = "database-network"
DOCKER_HOST_NIC_MODE = "docker-hostnic"
DOCKER_BRIDGE_MODE = "bridge"
MYSQL_HOST_SOCKET_PATH = "/var/lib/mysqld"
POSTGRESQL_HOST_SOCKET_PATH = "/var/lib/postgresql-socket"

REGISTRY_EXT_DEFAULTS = {
    'mysql':
        'trove.guestagent.datastore.mysql.manager.Manager',
    'mariadb':
        'trove.guestagent.datastore.mariadb.manager.Manager',
    'postgresql':
        'trove.guestagent.datastore.postgres.manager.PostgresManager',
    'percona':
        'trove.guestagent.datastore.experimental.percona.manager.Manager',
    'pxc':
        'trove.guestagent.datastore.experimental.pxc.manager.Manager',
    'redis':
        'trove.guestagent.datastore.experimental.redis.manager.Manager',
    'cassandra':
        'trove.guestagent.datastore.experimental.cassandra.manager.Manager',
    'couchbase':
        'trove.guestagent.datastore.experimental.couchbase.manager.Manager',
    'mongodb':
        'trove.guestagent.datastore.experimental.mongodb.manager.Manager',
    'couchdb':
        'trove.guestagent.datastore.experimental.couchdb.manager.Manager',
    'vertica':
        'trove.guestagent.datastore.experimental.vertica.manager.Manager',
    'db2':
        'trove.guestagent.datastore.experimental.db2.manager.Manager',
}



# Copyright 2020 Catalyst Cloud
# Copyright 2023 Bizfly Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_log import log as logging
from oslo_utils.excutils import save_and_reraise_exception

from trove.common import cfg
from trove.common import exception
from trove.common import utils
from trove.guestagent.common import operating_system
from trove.guestagent.datastore import service as base_service
from trove.guestagent.utils import docker as docker_utils


from trove.guestagent.datastore.mariadb import service
from trove.guestagent.datastore.mysql_common import manager


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class Manager(manager.MySqlManager):
    def __init__(self):
        status = base_service.BaseDbStatus(self.docker_client)
        app = service.MariaDBApp(status, self.docker_client)
        adm = service.MariaDBAdmin(app)

        super(Manager, self).__init__(app, status, adm)

    def get_start_db_params(self, data_dir):
        """Get parameters for starting database.

        Cinder volume initialization(after formatted) may leave a lost+found
        folder.
        """
        return (f'--ignore-db-dir=lost+found --ignore-db-dir=conf.d '
                f'--datadir={data_dir}')

    def pre_create_backup(self, context, **kwargs):
        LOG.info("Running pre_create_backup")
        status = {}
        try:
            INFO_FILE = "%s/xtrabackup_binlog_info" % self.app.get_data_dir()
            self.app.execute_sql("FLUSH TABLES WITH READ LOCK;")
            stt = self.app.execute_sql("SHOW MASTER STATUS;")
            for row in stt:
                status = {
                    'log_file': row._mapping['File'],
                    'log_position': row._mapping['Position']
                }

                for g in self.app.execute_sql(
                        "select @@global.gtid_current_pos;"):
                    gtid = g._mapping['@@global.gtid_current_pos']

                    status['log_executed_gtid_set'] = gtid

                binlog = "\t".join(map(str, [
                    status['log_file'],
                    status['log_position'],
                    status['log_executed_gtid_set']]))
                operating_system.write_file(INFO_FILE, binlog, as_root=True)

            mount_point = CONF.get(CONF.datastore_manager).mount_point
            operating_system.sync(mount_point)
            operating_system.fsfreeze(mount_point)
        except Exception as e:
            LOG.error("Run pre_create_backup failed, error: %s" % str(e))
            raise exception.BackupCreationError(str(e))
        return status

    def reset_password_for_restore(self, ds_version=None,
                                   data_dir='/var/lib/mysql/data'):
        """Reset the root password after restore the db data.

        uses --skip-grant-tables to temporarily disable auth and
        directly reset the root password via SQL.
        """
        LOG.info('Starting to reset password for restore')

        try:
            root_pass = self.app.get_auth_password(file="root.cnf")
        except exception.UnprocessableEntity:
            root_pass = utils.generate_random_password()
            self.app.save_password('root', root_pass)

        command = (
            f'--skip-grant-tables '
            f'--datadir={data_dir} '
        )
        reset_sql = (
            "FLUSH PRIVILEGES; "
            "ALTER USER 'root'@'localhost' IDENTIFIED BY '{}';"
        ).format(root_pass)
        reset_command = ["mariadb", "-u", "root", "-e", reset_sql]

        # Start the database container process.
        try:
            self.app.start_db(ds_version=ds_version, command=command)
            docker_utils.run_command(self.app.docker_client, reset_command)
        except Exception as err:
            with save_and_reraise_exception():
                LOG.error('Failed to reset password for restore, error: %s',
                          str(err))
        finally:
            try:
                LOG.debug(
                    'The init container log: %s',
                    docker_utils.get_container_logs(self.app.docker_client)
                )
                docker_utils.remove_container(self.app.docker_client)
            except Exception as err:
                LOG.error('Failed to remove container. error: %s',
                          str(err))
                pass
        LOG.info('Finished to reset password for restore')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


# NOTE(wuchunyang): these codes are copied from kuryr-libnetwork project.
EPSILON_PATTERN = '^$'
UUID_BASE = '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
UUID_PATTERN = EPSILON_PATTERN + '|' + UUID_BASE
IPV4_PATTERN_BASE = ('((25[0-5]|2[0-4][0-9]|1?[0-9]?[0-9])\\.){3}'
                     '(25[0-5]|2[0-4][0-9]|1?[0-9]?[0-9])')
CIDRV4_PATTERN = EPSILON_PATTERN + '|^(' + IPV4_PATTERN_BASE + \
    '(/(1[0-2][0-8]|[1-9]?[0-9]))' + ')$'
IPV6_PATTERN_BASE = ('('
                     '([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|'
                     '([0-9a-fA-F]{1,4}:){1,7}:|'
                     '([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|'
                     '([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|'
                     '([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|'
                     '([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|'
                     '([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|'
                     '[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|'
                     ':((:[0-9a-fA-F]{1,4}){1,7}|:)|'
                     'fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|'
                     '::(ffff(:0{1,4}){0,1}:){0,1}'
                     '((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\\.){3,3}'
                     '(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|'
                     '([0-9a-fA-F]{1,4}:){1,4}:'
                     '((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\\.){3,3}'
                     '(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))')
IPV6_PATTERN = EPSILON_PATTERN + '|^' + IPV6_PATTERN_BASE + '$'
CIDRV6_PATTERN = EPSILON_PATTERN + '|^(' + IPV6_PATTERN_BASE + \
    '(/(1[0-2][0-8]|[1-9]?[0-9]))' + ')$'

SCHEMA = {
    "PLUGIN_ACTIVATE": {"Implements": ["NetworkDriver"]},
    "SUCCESS": {}
}

COMMONS = {
    'description': 'Common data schemata shared among other schemata.',
    'links': [],
    'title': 'Trove Common Data Schema Definitions',
    'properties': {
        'options': {'$ref': '/schemata/commons#/definitions/options'},
        'mac': {'$ref': '/schemata/commons#/definitions/mac'},
        'cidrv6': {'$ref': '/schemata/commons#/definitions/cidrv6'},
        'interface': {'$ref': '/schemata/commons#/definitions/interface'},
        'cidr': {'$ref': '/schemata/commons#/definitions/cidr'},
        'id': {'$ref': '/schemata/commons#/definitions/id'},
        'uuid': {'$ref': '/schemata/commons#/definitions/uuid'},
        'ipv4': {'$ref': '/schemata/commons#/definitions/ipv4'},
    },
    'definitions': {
        'options': {
            'type': ['object', 'null'],
            'description': 'Options.',
            'example': {}
        },
        'id': {
            'oneOf': [
                {'pattern': '^([0-9a-f]{64})$'},
                {'pattern': '^([0-9a-z]{25})$'}],
            'type': 'string',
            'description': '64 or 25 length ID value of Docker.',
            'example': [
                '51c75a2515d47edecc3f720bb541e287224416fb66715eb7802011d6ffd4'
                '99f1',
                'xqqzd9p112o4kvok38n3caxjm'
            ]
        },
        'mac': {
            'pattern': (EPSILON_PATTERN + '|'
                        '^((?:[0-9a-f]{2}:){5}[0-9a-f]{2}|'
                        '(?:[0-9A-F]{2}:){5}[0-9A-F]{2})$'),
            'type': 'string',
            'description': 'A MAC address.',
            'example': 'aa:bb:cc:dd:ee:ff'
        },
        'cidr': {
            'pattern': CIDRV4_PATTERN,
            'type': 'string',
            'description': 'A IPv4 CIDR of the subnet.',
            'example': '10.0.0.0/24'
        },
        'cidrv6': {
            'pattern': CIDRV6_PATTERN,
            'type': 'string',
            'description': 'A IPv6 CIDR of the subnet.',
            'example': '10.0.0.0/24'
        },
        'ipv4datum': {
            'description': 'IPv4 data',
            'required': [
                'AddressSpace', 'Pool'],
            'type': 'object',
            'example': {
                'AddressSpace': 'foo',
                'Pool': '192.168.42.0/24',
                'Gateway': '192.168.42.1/24',
                'AuxAddresses': {
                    'web': '192.168.42.2',
                    'db': '192.168.42.3'
                }
            },
            'properties': {
                'AddressSpace': {
                    'description': 'The name of the address space.',
                    'type': 'string',
                    'example': 'foo',
                },
                'Pool': {
                    'description': 'A range of IP Addresses requested in '
                                   'CIDR format address/mask.',
                    '$ref': '#/definitions/commons/definitions/cidr'
                },
                'Gateway': {
                    'description': 'Optionally, the IPAM driver may provide '
                                   'a Gateway for the subnet represented by '
                                   'the Pool.',
                    '$ref': '#/definitions/commons/definitions/cidr',
                },
                'AuxAddresses': {
                    'description': 'A list of pre-allocated ip-addresses '
                                   'with an associated identifier as '
                                   'provided by the user to assist network '
                                   'driver if it requires specific '
                                   'ip-addresses for its operation.',
                    'type': 'object',
                    'patternProperties': {
                        '.+': {
                            'description': 'key-value pair of the ID and '
                                           'the IP address',
                            '$ref': '#/definitions/commons/definitions/ipv4'
                        }
                    }
                }
            }
        },
        'ipv6datum': {
            'description': 'IPv6 data',
            'required': [
                'AddressSpace', 'Pool', 'Gateway'],
            'type': 'object',
            'example': {
                'AddressCpace': 'bar',
                'Pool': 'fe80::/64',
                'Gateway': 'fe80::f816:3eff:fe20:57c3/64',
                'AuxAddresses': {
                    'web': 'fe80::f816:3eff:fe20:57c4',
                    'db': 'fe80::f816:3eff:fe20:57c5'
                }
            },
            'properties': {
                'AddressSpace': {
                    'description': 'The name of the address space.',
                    'type': 'string',
                    'example': 'foo',
                },
                'Pool': {
                    'description': 'A range of IP Addresses requested in '
                                   'CIDR format address/mask.',
                    '$ref': '#/definitions/commons/definitions/cidrv6'
                },
                'Gateway': {
                    'description': 'Optionally, the IPAM driver may provide '
                                   'a Gateway for the subnet represented by '
                                   'the Pool.',
                    '$ref': '#/definitions/commons/definitions/cidrv6',
                },
                'AuxAddresses': {
                    'description': 'A list of pre-allocated ip-addresses '
                                   'with an associated identifier as '
                                   'provided by the user to assist network '
                                   'driver if it requires specific '
                                   'ip-addresses for its operation.',
                    'type': 'object',
                    'patternProperties': {
                        '.+': {
                            'description': 'key-vavule pair of the ID and '
                                           'the IP address',
                            '$ref': '#/definitions/commons/definitions/ipv6'
                        }
                    }
                }
            }
        },
        'sandbox_key': {
            'pattern': '^(/var/run/docker/netns/[0-9a-f]{12})$',
            'type': 'string',
            'description': 'Sandbox information of netns.',
            'example': '/var/run/docker/netns/12bbda391ed0'
        },
        'uuid': {
            'pattern': UUID_PATTERN,
            'type': 'string',
            'description': 'uuid of neutron resources.',
            'example': 'dfe39822-ad5e-40bd-babd-3954113b3687'
        }
    },
    '$schema': 'https://json-schema.org/draft-04/hyper-schema',
    'type': 'object',
    'id': 'schemata/commons'
}

NETWORK_CREATE_SCHEMA = {
    'links': [{
        'method': 'POST',
        'href': '/NetworkDriver.CreateNetwork',
        'description': 'Create a Network',
        'rel': 'self',
        'title': 'Create'
    }],
    'title': 'Create network',
    'required': ['NetworkID', 'IPv4Data', 'IPv6Data', 'Options'],
    'definitions': {'commons': {}},
    '$schema': 'https://json-schema.org/draft-04/hyper-schema',
    'type': 'object',
    'properties': {
        'NetworkID': {
            'description': 'ID of a Network to be created',
            '$ref': '#/definitions/commons/definitions/id'
        },
        'IPv4Data': {
            'description': 'IPv4 data for the network',
            'type': 'array',
            'items': {
                '$ref': '#/definitions/commons/definitions/ipv4datum'
            }
        },
        'IPv6Data': {
            'description': 'IPv6 data for the network',
            'type': 'array',
            'items': {
                '$ref': '#/definitions/commons/definitions/ipv6datum'
            }
        },
        'Options': {
            'type': 'object',
            'description': 'Options',
            'required': ['com.docker.network.generic'],
            'properties': {
                'com.docker.network.generic': {
                    'type': 'object',
                    'required': ['hostnic_mac'],
                    'properties': {
                        'hostnic_mac': {
                            '$ref': '#/definitions/commons/definitions/mac'
                        }
                    }
                }
            }
        }
    }
}

NETWORK_CREATE_SCHEMA['definitions']['commons'] = COMMONS

NETWORK_JOIN_SCHEMA = {
    'links': [{
        'method': 'POST',
        'href': '/NetworkDriver.Join',
        'description': 'Join the network',
        'rel': 'self',
        'title': 'Create'
    }],
    'title': 'Join endpoint',
    'required': [
        'NetworkID',
        'EndpointID',
        'SandboxKey'
    ],
    'properties': {
        'NetworkID': {
            'description': 'Network ID',
            '$ref': '#/definitions/commons/definitions/id'
        },
        'SandboxKey': {
            'description': 'Sandbox Key',
            '$ref': '#/definitions/commons/definitions/sandbox_key'
        },
        'Options': {
            '$ref': '#/definitions/commons/definitions/options'
        },
        'EndpointID': {
            'description': 'Endpoint ID',
            '$ref': '#/definitions/commons/definitions/id'
        }
    },
    'definitions': {'commons': {}},
    '$schema': 'https://json-schema.org/draft-04/hyper-schema',
    'type': 'object',
}

NETWORK_JOIN_SCHEMA['definitions']['commons'] = COMMONS



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging

from trove.common import cfg
from trove.common import exception
from trove.guestagent.common import operating_system
from trove.guestagent.datastore.mysql import service
from trove.guestagent.datastore.mysql_common import manager
from trove.guestagent.datastore import service as base_service


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class Manager(manager.MySqlManager):
    def __init__(self):
        status = base_service.BaseDbStatus(self.docker_client)
        app = service.MySqlApp(status, self.docker_client)
        adm = service.MySqlAdmin(app)

        super(Manager, self).__init__(app, status, adm)

    def pre_create_backup(self, context, **kwargs):
        LOG.info("Running pre_create_backup")
        status = {}
        try:
            INFO_FILE = "%s/xtrabackup_binlog_info" % self.app.get_data_dir()
            self.app.execute_sql("FLUSH TABLES WITH READ LOCK;")
            if self.app._is_mysql84():
                stt = self.app.execute_sql("SHOW BINARY LOG STATUS;")
            else:
                stt = self.app.execute_sql("SHOW MASTER STATUS;")
            for row in stt:
                status = {
                    'log_file': row._mapping['File'],
                    'log_position': row._mapping['Position'],
                    'log_executed_gtid_set': row._mapping['Executed_Gtid_Set'],
                }

                binlog = "\t".join(map(str, [
                    status['log_file'],
                    status['log_position'],
                    status['log_executed_gtid_set']]))
                operating_system.write_file(INFO_FILE, binlog, as_root=True)

            mount_point = CONF.get(CONF.datastore_manager).mount_point
            operating_system.sync(mount_point)
            operating_system.fsfreeze(mount_point)
        except Exception as e:
            LOG.error("Run pre_create_backup failed, error: %s" % str(e))
            raise exception.BackupCreationError(str(e))

        return status



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.hacking import checks as tc
from trove.tests.unittests import trove_testtools


class HackingTestCase(trove_testtools.TestCase):

    def assertLinePasses(self, func, *args):
        def check_callable(f, *args):
            return next(f(*args))
        self.assertRaises(StopIteration, check_callable, func, *args)

    def assertLineFails(self, func, *args):
        self.assertIsInstance(next(func(*args)), tuple)

    def test_log_translations(self):
        all_log_levels = (
            'critical',
            'debug',
            'error',
            'exception',
            'info',
            'reserved',
            'warning',
        )
        for level in all_log_levels:
            bad = 'LOG.%s(_("Bad"))' % level
            self.assertEqual(
                1, len(list(tc.no_translate_logs(bad, 'f', False))))
            bad = "LOG.%s(_('Bad'))" % level
            self.assertEqual(
                1, len(list(tc.no_translate_logs(bad, 'f', False))))
            ok = 'LOG.%s("OK")' % level
            self.assertEqual(
                0, len(list(tc.no_translate_logs(ok, 'f', False))))
            ok = "LOG.%s(_('OK'))" % level
            self.assertEqual(
                0, len(list(tc.no_translate_logs(ok, 'f', True))))
            ok = "LOG.%s(variable)" % level
            self.assertEqual(
                0, len(list(tc.no_translate_logs(ok, 'f', False))))
            # Do not do validations in tests
            ok = 'LOG.%s(_("OK - unit tests"))' % level
            self.assertEqual(
                0, len(list(tc.no_translate_logs(ok, 'f/tests/f', False))))

    def test_check_localized_exception_messages(self):
        f = tc.check_raised_localized_exceptions
        self.assertLineFails(f, "     raise KeyError('Error text')", '')
        self.assertLineFails(f, ' raise KeyError("Error text")', '')
        self.assertLinePasses(f, ' raise KeyError(_("Error text"))', '')
        self.assertLinePasses(f, ' raise KeyError(_ERR("Error text"))', '')
        self.assertLinePasses(f, " raise KeyError(translated_msg)", '')
        self.assertLinePasses(f, '# raise KeyError("Not translated")', '')
        self.assertLinePasses(f, 'print("raise KeyError("Not '
                                 'translated")")', '')

    def test_check_localized_exception_message_skip_tests(self):
        f = tc.check_raised_localized_exceptions
        self.assertLinePasses(f, "raise KeyError('Error text')",
                              'neutron_lib/tests/unit/mytest.py')

    def test_no_log_warn(self):
        self.assertEqual(1, len(list(tc.no_log_warn(
            "LOG.warn(\"LOG.warn is deprecated\")"))))
        self.assertEqual(0, len(list(tc.no_log_warn(
            "LOG.warning(\"LOG.warn is deprecated\")"))))



# Copyright 2023 BizflyCloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from unittest import mock

from trove.common import cfg
from trove.common import exception
from trove.guestagent.datastore.mariadb import service
from trove.guestagent.datastore.mysql import service as mysql_service
from trove.guestagent.datastore import service as base_service
from trove.tests.unittests import trove_testtools


CONF = cfg.CONF


class TestService(trove_testtools.TestCase):
    def setUp(self):
        super(TestService, self).setUp()
        _docker_client = mock.MagicMock()
        status = mock.MagicMock()
        self.app = service.MariaDBApp(_docker_client, status)
        self.mysql_app = mysql_service.MySqlApp(_docker_client, status)

    def test_get_backup_image_not_defined(self):
        self.patch_datastore_manager('mariadb')
        self.assertRaises(exception.TroveError, self.app.get_backup_image)

    def test_get_backup_image_with_tag(self):
        self.patch_datastore_manager('mariadb')
        CONF.set_override('backup_docker_image',
                          'example.domain/repo/mariadb:tag', 'mariadb')
        image = self.app.get_backup_image()
        self.assertEqual(CONF.mariadb.backup_docker_image, image)

    def test_get_backup_image_without_tag(self):
        self.patch_datastore_manager('mariadb')
        CONF.set_override('backup_docker_image',
                          'example.domain/repo/mariadb', 'mariadb')
        self.patch_conf_property('datastore_version', '10.4')
        image = self.app.get_backup_image()
        _img = f'{CONF.mariadb.backup_docker_image}:{CONF.datastore_version}'
        self.assertEqual(_img, image)

    def test_mysql_backup_image_with_tag(self):
        self.patch_datastore_manager('mysql')
        CONF.set_override('backup_docker_image',
                          'example.domain/repo/mysql:1.1.0', 'mysql')
        self.patch_conf_property('datastore_version', '5.7')
        image = self.mysql_app.get_backup_image()
        self.assertEqual(image, "example.domain/repo/mysql5.7:1.1.0")

    def test_mysql_backup_image_without_tag(self):
        self.patch_datastore_manager('mysql')
        CONF.set_override('backup_docker_image',
                          'example.domain/repo/mysql', 'mysql')
        self.patch_conf_property('datastore_version', '5.7')
        image = self.mysql_app.get_backup_image()
        self.assertEqual(image, "example.domain/repo/mysql:5.7")

    def test_image_has_tag(self):
        fake_values = [
            "example.domain:5000/repo/image_name:tag",
            "example.domain:5000/repo/image-name:tag_tag",
            "example.domain:5000/repo/image_name:tag-tag",
            "example.domain:5000/repo/image-name",
            "example.domain:5000/repo/image_name",
            "example.domain/repo/image-name",
            "example.domain/repo/image-name:tag"]
        self.assertTrue(
            base_service.BaseDbApp._image_has_tag(fake_values[0]))
        self.assertTrue(
            base_service.BaseDbApp._image_has_tag(fake_values[1]))
        self.assertTrue(
            base_service.BaseDbApp._image_has_tag(fake_values[2]))
        self.assertFalse(
            base_service.BaseDbApp._image_has_tag(fake_values[3]))
        self.assertFalse(
            base_service.BaseDbApp._image_has_tag(fake_values[4]))
        self.assertFalse(
            base_service.BaseDbApp._image_has_tag(fake_values[5]))
        self.assertTrue(
            base_service.BaseDbApp._image_has_tag(fake_values[6]))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add Datastore Version Registry Extension


Revision ID: 5c68b4fb3cd1
Revises: 906cffda7b29
Create Date: 2024-04-30 13:59:10.690895
"""

from typing import Sequence, Union

from alembic import op
from sqlalchemy.sql import table, column
from sqlalchemy import text, Column
from sqlalchemy import String, Text

from trove.common.constants import REGISTRY_EXT_DEFAULTS

# revision identifiers, used by Alembic.
revision: str = '5c68b4fb3cd1'
down_revision: Union[str, None] = '906cffda7b29'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


repl_namespaces = {
    "mariadb": "trove.guestagent.strategies.replication.mariadb_gtid",
    "mongodb":
        "trove.guestagent.strategies.replication.experimental.mongo_impl",
    "mysql": "trove.guestagent.strategies.replication.mysql_gtid",
    "percona": "trove.guestagent.strategies.replication.mysql_gtid",
    "postgresql": "trove.guestagent.strategies.replication.postgresql",
    "pxc": "trove.guestagent.strategies.replication.mysql_gtid",
    "redis": "trove.guestagent.strategies.replication.experimental.redis_sync",

}

repl_strategies = {
    "mariadb": "MariaDBGTIDReplication",
    "mongodb": "Replication",
    "mysql": "MysqlGTIDReplication",
    "percona": "MysqlGTIDReplication",
    "postgresql": "PostgresqlReplicationStreaming",
    "pxc": "MysqlGTIDReplication",
    "redis": "RedisSyncReplication",

}


def upgrade() -> None:
    bind = op.get_bind()

    # 1. select id and manager from datastore_versions table
    connection = op.get_bind()
    # add columns before proceeding
    op.add_column("datastore_versions", Column('registry_ext', Text(),
                                               nullable=True))
    op.add_column("datastore_versions", Column('repl_strategy', Text(),
                                               nullable=True))
    for dsv_id, dsv_manager in connection.execute(
            text("select id, manager from datastore_versions")):
        registry_ext = REGISTRY_EXT_DEFAULTS.get(dsv_manager, '')
        repl_strategy = "%(repl_namespace)s.%(repl_strategy)s" % {
            'repl_namespace': repl_namespaces.get(dsv_manager, ''),
            'repl_strategy': repl_strategies.get(dsv_manager, '')
        }
        ds_versions_table = table(
            "datastore_versions",
            column("id", String),
            column("registry_ext", String),
            column("repl_strategy", String))
        op.execute(
            ds_versions_table.update()
            .where(ds_versions_table.c.id == dsv_id)
            .values({"registry_ext": registry_ext,
                     "repl_strategy": repl_strategy})
        )

    if bind.engine.name != "sqlite":
        op.alter_column("datastore_versions", "registry_ext", nullable=False,
                        existing_type=Text)
        op.alter_column("datastore_versions", "repl_strategy", nullable=False,
                        existing_type=Text)
    else:
        with op.batch_alter_table('datastore_versions') as bo:
            bo.alter_column("registry_ext", nullable=False,
                            existing_type=Text)
            bo.alter_column("repl_strategy", nullable=False,
                            existing_type=Text)


def downgrade() -> None:
    pass



# Copyright 2024 Bizfly Cloud, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from trove.common import cfg
from trove.common.clients import normalize_url
from trove.common import context
from trove.common import timeutils
from trove.common import trove_remote
from trove.tests.unittests import trove_testtools

CONF = cfg.CONF


class TestTroveClient(trove_testtools.TestCase):
    def setUp(self):
        super(TestTroveClient, self).setUp()
        self.context = context.TroveContext(
            project_id='TENANT-' + str(timeutils.utcnow()),
            user_id='USER-' + str(timeutils.utcnow()),
            auth_token='AUTH_TOKEN-' + str(timeutils.utcnow()))

    def test_trove_with_remote_client(self):
        self.patch_conf_property('trove_url', 'trove_url')
        client = trove_remote.trove_client(self.context)
        url = '%(url)s%(tenant)s' % {
            'url': normalize_url(CONF.trove_url),
            'tenant': self.context.project_id}
        self.assertEqual(url, client.client.management_url)



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
import re

import docker
from docker import errors as derros
from docker import types
from oslo_log import log as logging
from oslo_utils import encodeutils

from trove.common import cfg
from trove.common import constants

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
ANSI_ESCAPE = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')


def stop_container(client, name="database"):
    try:
        container = client.containers.get(name)
    except docker.errors.NotFound:
        LOG.warning("Failed to get container %s", name)
        return

    container.stop(timeout=CONF.state_change_wait_time)


def create_network(client: docker.client.DockerClient,
                   name: str) -> str:
    networks = client.networks.list()
    for net in networks:
        if net.name == name:
            return net.id
    LOG.debug("Creating docker network: %s", name)
    with open(constants.ETH1_CONFIG_PATH) as fd:
        eth1_config = json.load(fd)
    enable_ipv6 = False
    ipam_pool = list()
    if eth1_config.get("ipv4_address"):
        ipam_pool.append(types.IPAMPool(
            subnet=eth1_config.get("ipv4_cidr"),
            gateway=eth1_config.get("ipv4_gateway"))
        )
    if eth1_config.get("ipv6_address"):
        enable_ipv6 = True
        ipam_pool.append(types.IPAMPool(
            subnet=eth1_config.get("ipv6_cidr"),
            gateway=eth1_config.get("ipv6_gateway")
        ))

    ipam_config = docker.types.IPAMConfig(pool_configs=ipam_pool)
    mac_address = eth1_config.get("mac_address")
    net = client.networks.create(name=name,
                                 driver=constants.DOCKER_HOST_NIC_MODE,
                                 ipam=ipam_config,
                                 enable_ipv6=enable_ipv6,
                                 options=dict(hostnic_mac=mac_address))
    LOG.debug("docker network: %s created successfully", net.id)
    return net.id


def _create_container_with_low_level_api(image: str, param: dict) -> None:
    # create a low-level docker api object
    client = docker.APIClient(base_url='unix://var/run/docker.sock')
    host_config_kwargs = dict()
    if param.get("restart_policy"):
        host_config_kwargs["restart_policy"] = param.get("restart_policy")
    if param.get("privileged"):
        host_config_kwargs["privileged"] = param.get("privileged")
    if param.get("volumes"):
        host_config_kwargs["binds"] = param.get("volumes")
    host_config = client.create_host_config(**host_config_kwargs)

    network_config_kwargs = dict()
    with open(constants.ETH1_CONFIG_PATH) as fd:
        eth1_config = json.load(fd)
    if eth1_config.get("ipv4_address"):
        network_config_kwargs["ipv4_address"] = eth1_config.get("ipv4_address")
    if eth1_config.get("ipv6_address"):
        network_config_kwargs["ipv6_address"] = eth1_config.get("ipv6_address")

    networking_config = client.create_networking_config(
        {param.get("network"):
            client.create_endpoint_config(**network_config_kwargs)})
    healthcheck = param.get("healthcheck", None)

    # NOTE(wuchunyang): the low-level api doesn't support RUN interface,
    # so we need pull image first, then start the container
    LOG.debug("Pulling docker images: %s", image)
    try:
        client.pull(image)
    except derros.APIError as e:
        LOG.error("failed to pull image: %s, due to the error: %s", image, e)
        raise
    LOG.debug("Creating container: %s", param.get("name"))
    container = client.create_container(image=image,
                                        name=param.get("name"),
                                        detach=param.get("detach"),
                                        user=param.get("user"),
                                        environment=param.get("environment"),
                                        command=param.get("command"),
                                        host_config=host_config,
                                        healthcheck=healthcheck,
                                        networking_config=networking_config)
    LOG.debug("Starting container: %s", param.get("name"))
    client.start(container=container)


def start_container(client, image, name="database",
                    restart_policy="unless-stopped",
                    volumes={}, ports={}, user="", network_mode="host",
                    environment={}, command="", healthcheck=None,
                    privileged=False):
    """Start a docker container.

    :param client: docker client obj.
    :param image: docker image.
    :param name: container name.
    :param restart_policy: restart policy.
    :param volumes: e.g.
           {"/host/trove": {"bind": "/container/trove", "mode": "rw"}}
    :param ports: ports is ignored when network_mode="host". e.g.
           {"3306/tcp": 3306}
    :param user: e.g. "1000.1001"
    :param network_mode: One of bridge, none, host
    :param environment: Environment variables
    :param command:
    :param privileged: docker privileged
    :return:
    """
    try:
        container = client.containers.get(name)
        LOG.info(f'Starting existing container {name}')
        container.start()
        return
    except docker.errors.NotFound:
        pass

    LOG.info(
        f"Creating docker container, image: {image}, "
        f"volumes: {volumes}, ports: {ports}, user: {user}, "
        f"network_mode: {network_mode}, environment: {environment}, "
        f"health_check: {healthcheck}, "
        f"command: {command}")
    kwargs = dict(name=name,
                  restart_policy={"Name": restart_policy},
                  privileged=privileged,
                  detach=True,
                  volumes=volumes,
                  ports=ports,
                  user=user,
                  environment=environment,
                  healthcheck=healthcheck,
                  command=command)
    if network_mode == constants.DOCKER_HOST_NIC_MODE:
        create_network(client, constants.DOCKER_NETWORK_NAME)
        kwargs["network"] = constants.DOCKER_NETWORK_NAME
        return _create_container_with_low_level_api(image, kwargs)
    else:
        kwargs["network_mode"] = network_mode
        return client.containers.run(image, **kwargs)


def _decode_output(output):
    output = encodeutils.safe_decode(output)
    output = ANSI_ESCAPE.sub('', output.strip())
    return output.split('\n')


def run_container(client, image, name, network_mode="host", volumes={},
                  command="", user=""):
    """Run command in a container and return the string output list.

    :returns output: The log output.
    :returns ret: True if no error occurs, otherwise False.
    """
    try:
        container = client.containers.get(name)
        LOG.debug(f'Removing existing container {name}')
        container.remove(force=True)
    except docker.errors.NotFound:
        pass

    try:
        LOG.info(
            f'Running container {name}, image: {image}, '
            f'network_mode: {network_mode}, volumes: {volumes}, '
            f'command: {command}')
        output = client.containers.run(
            image,
            name=name,
            network_mode=network_mode,
            volumes=volumes,
            remove=False,
            command=command,
            user=user,
        )
    except docker.errors.ContainerError as err:
        output = err.container.logs()
        return _decode_output(output), False

    return _decode_output(output), True


def get_container_status(client, name="database"):
    try:
        container = client.containers.get(name)
        # One of created, restarting, running, removing, paused, exited, or
        # dead
        return container.status
    except docker.errors.NotFound:
        return "not running"
    except Exception:
        return "unknown"


def get_container_health(client: docker.DockerClient, name="database") -> str:
    try:
        container = client.containers.get(name)
        # One of staring, healthy, unhealthy, unknown
        return container.health
    except docker.errors.NotFound:
        return "not running"
    except Exception:
        return "unknown"


def run_command(client, command, name="database"):
    container = client.containers.get(name)
    # output is Bytes type
    ret, output = container.exec_run(command)
    if ret == 1:
        raise Exception('Running command error: %s' % output)

    return output


def restart_container(client, name="database"):
    container = client.containers.get(name)
    container.restart(timeout=CONF.state_change_wait_time)


def remove_container(client, name="database"):
    try:
        container = client.containers.get(name)
        container.remove(force=True)
    except docker.errors.NotFound:
        pass


def get_container_logs(client, name='database', tail=50):
    container = client.containers.get(name)
    output = container.logs(tail=tail)
    return _decode_output(output)


def prune_images(client):
    """Remove unused images."""
    try:
        client.images.prune(filters={'dangling': False})
    except Exception as e:
        LOG.warning(f"Prune image failed, error: {str(e)}")



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pathlib import Path


from alembic import command as alembic_command
from alembic import config as alembic_config
from alembic import migration as alembic_migration
from alembic.script import ScriptDirectory
from oslo_log import log as logging
import sqlalchemy as sa
import sqlalchemy.exc
from sqlalchemy import text

from trove.common import exception
from trove.db.sqlalchemy import session

LOG = logging.getLogger(__name__)

ALEMBIC_INIT_VERSION = '906cffda7b29'
ALEMBIC_LATEST_VERSION = '5c68b4fb3cd1'


def list(query_func, *args, **kwargs):
    query = query_func(*args, **kwargs)
    res = query.all()
    query.session.commit()
    return res


def count(query, *args, **kwargs):
    query = query(*args, **kwargs)
    res = query.count()
    query.session.commit()
    return res


def first(query, *args, **kwargs):
    query = query(*args, **kwargs)
    res = query.first()
    query.session.commit()
    return res


def join(query, model, *args):
    query = query(model)
    res = query.join(*args)
    query.session.commit()
    return res


def find_all(model, **conditions):
    return _query_by(model, **conditions)


def find_all_by_limit(query_func, model, conditions, limit, marker=None,
                      marker_column=None):
    query = _limits(query_func, model, conditions, limit, marker,
                    marker_column)
    res = query.all()
    query.session.commit()
    return res


def find_by(model, **kwargs):
    query = _query_by(model, **kwargs)
    res = query.first()
    query.session.commit()
    return res


def find_by_filter(model, **kwargs):
    filters = kwargs.pop('filters', [])
    return _query_by_filter(model, *filters, **kwargs)


def save(model):
    try:
        db_session = session.get_session()
        with db_session.begin():
            model = db_session.merge(model)
            db_session.flush()
            return model
    except sqlalchemy.exc.IntegrityError as error:
        raise exception.DBConstraintError(model_name=model.__class__.__name__,
                                          error=str(error.orig))


def delete(model):
    db_session = session.get_session()
    with db_session.begin():
        model = db_session.merge(model)
        db_session.delete(model)
        db_session.flush()


def delete_all(query_func, model, **conditions):
    query = query_func(model, **conditions)
    query.delete()
    query.session.commit()


def update(model, **values):
    for k, v in values.items():
        model[k] = v


def update_all(query_func, model, conditions, values):
    query = query_func(model, **conditions)
    query.update()
    query.session.commit()


def configure_db(*plugins):
    session.configure_db()
    configure_db_for_plugins(*plugins)


def configure_db_for_plugins(*plugins):
    for plugin in plugins:
        session.configure_db(models_mapper=plugin.mapper)


def drop_db():
    session.drop_db()


def clean_db():
    session.clean_db()


def _get_alembic_revision(config):
    script = ScriptDirectory.from_config(config)
    current_revision = script.get_current_head()
    if current_revision is not None:
        return current_revision
    return "head"


def _migrate_legacy_database(config):
    """Check if database is a legacy sqlalchemy-migrate-managed database.

    If it is, migrate it by "stamping" the initial alembic schema.
    """
    # If the database doesn't have the sqlalchemy-migrate legacy migration
    # table, we don't have anything to do
    engine = session.get_engine()
    if not sa.inspect(engine).has_table('migrate_version'):
        return

    # Likewise, if we've already migrated to alembic, we don't have anything to
    # do
    with engine.begin() as connection:
        context = alembic_migration.MigrationContext.configure(connection)
        if context.get_current_revision():
            return

    # We have legacy migrations but no alembic migration. Stamp (dummy apply)
    # the initial alembic migration.

    LOG.info(
        'The database is still under sqlalchemy-migrate control; '
        'fake applying the initial alembic migration'
    )
    # In case we upgrade from the branch prior to stable/2023.2
    if sa.inspect(engine).has_table('migrate_version'):
        # for the deployment prior to Bobocat
        query = text("SELECT version FROM migrate_version")
        with engine.connect() as connection:
            result = connection.execute(query)
        cur_version = result.first().values()[0]
        LOG.info("current version is %s", cur_version)
        if cur_version == 48:
            alembic_command.stamp(config, ALEMBIC_INIT_VERSION)
        elif cur_version > 48:
            # we already upgrade to the latest branch, use the latest
            # version(5c68b4fb3cd1)
            alembic_command.stamp(config, ALEMBIC_LATEST_VERSION)
        else:
            message = ("You need to upgrade trove database to a version "
                       "between Wallaby and Bobocat, and then upgrade to "
                       "the latest.")
            raise exception.BadRequest(message)


def _configure_alembic(conf=None):
    if conf is None:
        conf = conf.CONF
    alembic_ini = Path(__file__).joinpath('..', 'alembic.ini').resolve()
    if alembic_ini.exists():
        # alembic configuration
        config = alembic_config.Config(alembic_ini)
        # override the database configuration from the file
        config.set_main_option('sqlalchemy.url',
                               conf['database']['connection'])
        # override the logger configuration from the file
        # https://stackoverflow.com/a/42691781/613428
        config.attributes['configure_logger'] = False
        return config
    else:
        # return None if no alembic.ini exists
        return None


def db_sync(conf=None, version=None, repo_path=None):
    config = _configure_alembic(conf=conf)
    if config:
        # Check the version
        if version is None:
            version = _get_alembic_revision(config)
        # Raise an exception in sqlalchemy-migrate style
        if version is not None and version.isdigit():
            raise exception.InvalidValue(
                'You requested an sqlalchemy-migrate database version;'
                'this is no longer supported'
            )
        # Upgrade to a later version using alembic
        _migrate_legacy_database(config)
        alembic_command.upgrade(config, version)
    else:
        raise exception.BadRequest('sqlalchemy-migrate is '
                                   'no longer supported')


def db_upgrade(conf=None, version=None, repo_path=None):
    config = _configure_alembic(conf=conf)
    if config:
        # Check the version
        if version is None:
            version = 'head'
        # Raise an exception in sqlalchemy-migrate style
        if version.isdigit():
            raise exception.InvalidValue(
                'You requested an sqlalchemy-migrate database version;'
                'this is no longer supported'
            )
        # Upgrade to a later version using alembic
        _migrate_legacy_database(config)
        alembic_command.upgrade(config, version)
    else:
        raise exception.BadRequest('sqlalchemy-migrate is '
                                   'no longer supported')


def db_reset(*plugins):
    drop_db()
    db_sync()
    configure_db()


def _base_query(cls):
    db_session = session.get_session()
    query = db_session.query(cls)
    return query


def _query_by(cls, **conditions):
    query = _base_query(cls)
    if conditions:
        query = query.filter_by(**conditions)
    return query


def _query_by_filter(cls, *filters, **conditions):
    query = _query_by(cls, **conditions)
    if filters:
        query = query.filter(*filters)
    return query


def _limits(query_func, model, conditions, limit, marker, marker_column=None):
    query = query_func(model, **conditions)
    marker_column = marker_column or model.id
    if marker:
        query = query.filter(marker_column > marker)
    return query.order_by(marker_column).limit(limit)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""drop-migrate-version-table

Revision ID: cee1bcba3541
Revises: 5c68b4fb3cd1
Create Date: 2024-06-05 14:27:15.530991

"""
from typing import Sequence, Union

from alembic import op
from sqlalchemy.engine import reflection


# revision identifiers, used by Alembic.
revision: str = 'cee1bcba3541'
down_revision: Union[str, None] = '5c68b4fb3cd1'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    conn = op.get_bind()
    inspector = reflection.Inspector.from_engine(conn)
    tables = inspector.get_table_names()
    if 'migrate_version' in tables:
        op.drop_table('migrate_version')


def downgrade() -> None:
    pass
